import os
import os.path
from dataclasses import dataclass, field
from typing import Optional, List, Generator, Dict, Callable, Any

from ..php.parsing import parse_php_file, PhpException, PhpState, \
    PhpEvaluationOptions
from ..logging import log
from ..util.io import is_symlink_loop, PathSet, resolve_path, \
    resolve_parent_path
from .exceptions import WordpressException, ExtensionException
from .plugin import Plugin, PluginLoader
from .theme import Theme, ThemeLoader
from .database import WordpressDatabase, WordpressDatabaseServer, \
    DEFAULT_PORT, DEFAULT_COLLATION

WP_BLOG_HEADER_NAME = b'wp-blog-header.php'
WP_CONFIG_NAME = b'wp-config.php'

EXPECTED_CORE_FILES = {
        WP_BLOG_HEADER_NAME,
        b'wp-load.php'
    }
EXPECTED_CORE_DIRECTORIES = {
        b'wp-admin',
        b'wp-includes'
    }

EVALUATION_OPTIONS = PhpEvaluationOptions(
        allow_includes=False
    )

ALTERNATE_RELATIVE_CONTENT_PATHS = [
        b'../app'
    ]

DATABASE_CONFIG_CONSTANTS = {
        b'DB_NAME': 'name',
        b'DB_USER': 'user',
        b'DB_PASSWORD': 'password',
        b'DB_HOST': 'host',
        b'DB_COLLATE': 'collation'
    }


@dataclass
class WordpressStructureOptions:
    relative_content_paths: List[str] = field(default_factory=list)
    relative_plugins_paths: List[str] = field(default_factory=list)
    relative_mu_plugins_paths: List[str] = field(default_factory=list)


class PathResolver:

    def __init__(self, path: bytes):
        self.path = path

    def _resolve_path(self, path: bytes, base: bytes) -> bytes:
        return os.path.join(base, path.lstrip(b'/'))

    def resolve_path(self, path: bytes) -> bytes:
        return self._resolve_path(path, self.path)


class WordpressLocator(PathResolver):

    def __init__(
                self,
                path: bytes,
                allow_nested: bool = True,
                allow_io_errors: bool = False
            ):
        super().__init__(path)
        self.allow_nested = allow_nested
        self.allow_io_errors = allow_io_errors

    def _is_core_directory(self, path: bytes, quiet: bool = False) -> bool:
        missing_files = EXPECTED_CORE_FILES.copy()
        missing_directories = EXPECTED_CORE_DIRECTORIES.copy()
        try:
            for file in os.scandir(path):
                try:
                    if file.is_file():
                        if file.name in missing_files:
                            missing_files.remove(file.name)
                    elif file.is_dir():
                        if file.name in missing_directories:
                            missing_directories.remove(file.name)
                except OSError as error:
                    if self.allow_io_errors:
                        log.warning(
                                'Unable to determine if ' +
                                os.fsdecode(file.path) + ' is an '
                                'expected WordPress file as its type could '
                                f'not be determined: {error}'
                            )
                        continue
                    else:
                        raise
            if len(missing_files) > 0 or len(missing_directories) > 0:
                return False
            return True
        except OSError as error:
            if self.allow_io_errors:
                log.warning(
                        'Unable to scan directory at ' + os.fsdecode(path)
                        + f': {error}'
                    )
                return False
            else:
                raise WordpressException(
                        'Unable to scan directory at ' + os.fsdecode(path)
                    ) from error
        return False

    def _extract_core_path_from_index(self) -> Optional[str]:
        try:
            context = parse_php_file(self.resolve_path(b'index.php'))
            for include in context.get_includes():
                path = include.evaluate_path(context.state)
                basename = os.path.basename(path)
                if basename == WP_BLOG_HEADER_NAME:
                    return os.path.dirname(path)
        except PhpException:
            # If parsing fails, it's not a valid WordPress index file
            pass
        return None

    def _get_child_directories(
                self,
                path: bytes,
                processed: PathSet
            ) -> List[bytes]:
        directories = []
        for file in os.scandir(path):
            try:
                if file.is_dir():
                    if file.is_symlink() and \
                            is_symlink_loop(file.path, processed):
                        continue
                    directories.append(os.path.realpath(file.path))
            except OSError as error:
                if self.allow_io_errors:
                    log.warning(
                            'Ignoring child entry at '
                            + os.fsdecode(file.path) + ' as its type '
                            f'could not be determined: {error}'
                        )
                else:
                    raise WordpressException(
                            'Unable to determine type of file at '
                            + os.fsdecode(file.path)
                        )
        return directories

    def _search_for_core_directory(
                self,
                located: PathSet,
                processed: PathSet
            ) -> Generator[bytes, None, None]:
        paths = [self.path]
        while len(paths) > 0:
            directories = set()
            for path in paths:
                try:
                    directories.update(
                            self._get_child_directories(path, processed)
                        )
                except OSError as error:
                    message = (
                            'Unable to search child directory at '
                            + os.fsdecode(path) + ' due to IO error'
                        )
                    if self.allow_io_errors:
                        log.warning(message + f': {error}')
                    else:
                        raise WordpressException(message) from error
            paths = set()
            for directory in directories:
                processed.add(directory)
                if self._is_core_directory(directory):
                    if directory not in located:
                        yield directory
                        if self.allow_nested:
                            paths.add(directory)
                        located.add(directory)
                else:
                    paths.add(directory)

    def locate_core_paths(self) -> Generator[bytes, None, None]:
        located = PathSet()
        if self._is_core_directory(self.path):
            yield self.path
            if not self.allow_nested:
                return
            located.add(resolve_path(self.path))
        path = self._extract_core_path_from_index()
        if path is None:
            processed = PathSet()
            processed.add(self.path)
            yield from self._search_for_core_directory(located, processed)
        else:
            yield os.fsencode(path)

    def locate_parent_installation(self) -> Optional[bytes]:
        current = resolve_path(self.path)
        if not os.path.isdir(current):
            current = resolve_parent_path(current)
        while True:
            if self._is_core_directory(current):
                return current
            parent = resolve_parent_path(current)
            if parent == current:
                break
            current = parent
        return None


def locate_core_path(
            path: bytes,
            up: bool = False,
            allow_io_errors: bool = False
        ) -> bytes:
    locator = WordpressLocator(path, allow_io_errors=allow_io_errors)
    if up:
        core_path = locator.locate_parent_installation()
        if core_path is None:
            raise WordpressException(
                    'Unable to locate core files above '
                    + os.fsdecode(path)
                )
        return core_path
    else:
        for path in locator.locate_core_paths():
            return path
        raise WordpressException(
                'Unable to locate core files under '
                + os.fsdecode(path)
            )


class WordpressSite(PathResolver):

    def __init__(
                self,
                path: bytes,
                structure_options: Optional[WordpressStructureOptions] = None,
                core_path: bytes = None,
                is_child_path: bool = False,
                allow_io_errors: bool = False
            ):
        super().__init__(path)
        self.core_path = locate_core_path(path, is_child_path, allow_io_errors)
        self.structure_options = structure_options \
            if structure_options is not None else WordpressStructureOptions()
        self._version = None

    def resolve_core_path(self, path: bytes) -> bytes:
        return self._resolve_path(path, self.core_path)

    def resolve_content_path(self, path: bytes) -> bytes:
        return self._resolve_path(path, self.get_content_directory())

    def _determine_version(self) -> bytes:
        version_path = self.resolve_core_path(b'wp-includes/version.php')
        context = parse_php_file(version_path)
        try:
            state = context.evaluate(
                    options=EVALUATION_OPTIONS
                )
            version = state.get_variable_value(b'wp_version')
            if isinstance(version, bytes):
                return version
        except PhpException as exception:
            raise WordpressException(
                    'Unable to parse WordPress version file at '
                    + os.fsdecode(version_path)
                ) from exception
        raise WordpressException('Unable to determine WordPress version')

    def get_version(self) -> str:
        if self._version is None:
            self._version = self._determine_version()
        return self._version

    def _locate_config_file(self) -> str:
        paths = [
                self.resolve_core_path(b'wp-config.php'),
                os.path.join(os.path.dirname(self.core_path), b'wp-config.php')
            ]
        for path in paths:
            if os.path.isfile(path):
                return path
        return None

    def _parse_config_file(self) -> Optional[PhpState]:
        config_path = self._locate_config_file()
        try:
            if config_path is not None:
                context = parse_php_file(config_path)
                return context.evaluate(
                        options=EVALUATION_OPTIONS
                    )
        except PhpException as exception:
            # Ignore config files that cannot be parsed
            log.debug(
                    'Unable to parse WordPress config file at '
                    + os.fsdecode(config_path)
                    + f' : {exception}'
                )
        return None

    def _get_parsed_config_state(self) -> PhpState:
        if not hasattr(self, 'config_state'):
            self.config_state = self._parse_config_file()
        return self.config_state

    def _extract_string_from_config(
                self,
                constant: bytes,
                default: Optional[bytes],
                extractor: Callable[[PhpState], Any]
            ) -> bytes:
        try:
            state = self._get_parsed_config_state()
            if state is not None:
                value = extractor(state)
                if isinstance(value, bytes):
                    return value
        except PhpException as exception:
            # Just use the default if parsing errors occur
            log.warning(
                    f'Unable to extract constant {constant} from WordPress '
                    f'config: {exception}'
                )
        return default

    def _extract_string_from_config_constant(
                self,
                constant: bytes,
                default: Optional[bytes] = None
            ):
        def get_constant_value(state: PhpState):
            return state.get_constant_value(
                    name=constant,
                    default_to_name=False
                )
        return self._extract_string_from_config(
                constant,
                default,
                get_constant_value
            )

    def _extract_string_from_config_variable(
                self,
                variable: bytes,
                default: Optional[bytes] = None
            ):
        def get_variable_value(state: PhpState):
            return state.get_variable_value(variable)
        return self._extract_string_from_config(
                variable,
                default,
                get_variable_value
            )

    def get_config_constant(self, constant: bytes) -> bytes:
        return self._extract_string_from_config_constant(constant)

    def get_config_variable(self, variable: bytes) -> bytes:
        return self._extract_string_from_config_variable(variable)

    def _generate_possible_content_paths(self) -> Generator[str, None, None]:
        configured = self._extract_string_from_config_constant(
                'WP_CONTENT_DIR'
            )
        if configured is not None:
            yield configured
        for path in self.structure_options.relative_content_paths:
            yield self.resolve_core_path(path)
        for path in ALTERNATE_RELATIVE_CONTENT_PATHS:
            yield self.resolve_core_path(path)
        yield self.resolve_core_path(b'wp-content')

    def _locate_content_directory(self) -> str:
        for path in self._generate_possible_content_paths():
            log.debug('Checking potential content path: ' + os.fsdecode(path))
            possible_themes_path = self._resolve_path(b'themes', path)
            if os.path.isdir(path) and os.path.isdir(possible_themes_path):
                log.debug('Located content directory at ' + os.fsdecode(path))
                return path
        raise WordpressException(
                'Unable to locate content directory for site at '
                + os.fsdecode(self.path)
            )

    def get_content_directory(self) -> str:
        if not hasattr(self, 'content_path'):
            self.content_path = self._locate_content_directory()
        return self.content_path

    def get_configured_plugins_directory(self, mu: bool = False) -> str:
        return self._extract_string_from_config_constant(
                'WPMU_PLUGIN_DIR' if mu else 'WP_PLUGIN_DIR',
            )

    def _generate_possible_plugins_paths(
                self,
                mu: bool = False,
                allow_io_errors: bool = False
            ) -> Generator[str, None, None]:
        configured = self.get_configured_plugins_directory(mu)
        if configured is not None:
            yield configured
        relative_paths = self.structure_options.relative_mu_plugins_paths \
            if mu else self.structure_options.relative_plugins_paths
        for path in relative_paths:
            yield self.resolve_core_path(path)
        try:
            yield self.resolve_content_path(
                    b'mu-plugins' if mu else b'plugins'
                )
        except WordpressException:
            if not allow_io_errors:
                raise

    def get_plugins(
                self,
                mu: bool = False,
                allow_io_errors: bool = False
            ) -> List[Plugin]:
        log_plugins = 'must-use plugins' if mu else 'plugins'
        for path in self._generate_possible_plugins_paths(mu, allow_io_errors):
            log.debug(
                    f'Checking potential {log_plugins} path: '
                    + os.fsdecode(path)
                )
            loader = PluginLoader(path, allow_io_errors)
            try:
                plugins = loader.load_all()
                log.debug(
                        f'Located {log_plugins} directory at '
                        + os.fsdecode(path)
                    )
                return plugins
            except ExtensionException:
                # If extensions can't be loaded, the directory is not valid
                continue
        if mu:
            log.warning(
                    'No mu-plugins directory found for site at '
                    + os.fsdecode(self.path)
                )
            return []
        if allow_io_errors:
            return []
        raise WordpressException(
                f'Unable to locate {log_plugins} directory for site at '
                + os.fsdecode(self.path)
            )

    def get_all_plugins(self, allow_io_errors: bool = False) -> List[Plugin]:
        plugins = self.get_plugins(mu=True, allow_io_errors=allow_io_errors)
        plugins += self.get_plugins(mu=False, allow_io_errors=allow_io_errors)
        return plugins

    def get_theme_directory(self) -> str:
        return self.resolve_content_path(b'themes')

    def get_themes(self, allow_io_errors: bool = False) -> List[Theme]:
        try:
            directory = self.get_theme_directory()
        except WordpressException:
            if allow_io_errors:
                return []
            else:
                raise
        loader = ThemeLoader(directory, allow_io_errors)
        return loader.load_all()

    def _extract_database_config(self) -> Dict[str, str]:
        config = {}

        def add_config(key: str, value: Any):
            if value is None:
                raise WordpressException(
                        'Unable to extract database connection details from '
                        f'WordPress config (Key: {key}, Value: '
                        + repr(value) + ')'
                    )
            config[key] = value.decode('latin1')

        for constant, attribute in DATABASE_CONFIG_CONSTANTS.items():
            add_config(
                    key=attribute,
                    value=self.get_config_constant(constant)
                )
        add_config(
                key='prefix',
                value=self.get_config_variable(b'table_prefix')
            )
        return config

    def get_database(self) -> WordpressDatabase:
        config = self._extract_database_config()
        host_components = config['host'].split(':', 1)
        host = host_components[0]
        try:
            port = int(host_components[1])
        except IndexError:
            port = DEFAULT_PORT
        try:
            collation = config['collation']
        except KeyError:
            collation = DEFAULT_COLLATION
        server = WordpressDatabaseServer(
                host=host,
                port=port,
                user=config['user'],
                password=config['password']
            )
        return WordpressDatabase(
                name=config['name'],
                server=server,
                prefix=config['prefix'],
                collation=collation
            )



import sys
from typing import List, Iterable, Dict, Optional

from .config import resolve_config_map
from .config.config_items import ConfigItemDefinition
from .subcommands import SubcommandDefinition, load_subcommand_definition, \
        VALID_SUBCOMMANDS


def _write_bool(value: bool) -> None:
    print('true' if value else '')


def _write_options(
            options: Iterable[str],
            allow_files: bool = False,
            allow_directories: bool = False
        ) -> None:
    print(' '.join(options))
    _write_bool(allow_files)
    _write_bool(allow_directories)


def _write_completion_options(
            words: List[str],
            subcommand_definition: SubcommandDefinition,
            config_map: Dict[str, ConfigItemDefinition],
            previous: Optional[str] = None
        ) -> bool:
    options = []
    allow_files = subcommand_definition.accepts_files
    allow_directories = subcommand_definition.accepts_directories
    if subcommand_definition.name == 'help':
        options.extend(VALID_SUBCOMMANDS)
    for item in config_map.values():
        item_options = {
                f'--{item.name}'
            }
        if item.short_name is not None:
            item_options.add(f'-{item.short_name}')
        if item.is_flag():
            item_options.add(f'--no-{item.name}')
        if previous in item_options and item.accepts_value():
            options = [] if item.meta.valid_options is None \
                else list(item.meta.valid_options)
            allow_files = item.meta.accepts_file
            allow_directories = item.meta.accepts_directory
            break
        options.extend(item_options)
    _write_options(
            options,
            allow_files,
            allow_directories
        )


def auto_complete(words: List[str], cursor_index: int) -> None:
    try:
        subcommand = words[1]
    except IndexError:
        subcommand = None
    # cursor_word = words[cursor_index]
    if cursor_index == 1 or subcommand is None:
        _write_options(VALID_SUBCOMMANDS)
    else:
        subcommand_definition = load_subcommand_definition(subcommand)
        config_map = resolve_config_map(subcommand_definition)
        try:
            previous = words[cursor_index - 1]
        except IndexError:
            previous = None
        _write_completion_options(
                words,
                subcommand_definition,
                config_map,
                previous
            )


if __name__ == '__main__':
    words = sys.argv[1:-1]
    index = sys.argv[-1]
    if not index.isdecimal():
        raise Exception('Cursor index must be a number')
    index = int(index)
    auto_complete(words, index)



import re
import os
from typing import Optional, List, Callable, Pattern, AnyStr


class FilterCondition:

    def __init__(self, test: Callable[[str], bool], allow: bool = True):
        self.test = test
        self.allow = allow

    def evaluate(self, path: bytes) -> bool:
        return self.test(path)


class FileFilter:

    def __init__(self, conditions: Optional[List[FilterCondition]] = None):
        self._conditions = conditions if conditions is not None else []

    def add_condition(self, condition: FilterCondition) -> None:
        self._conditions.append(condition)

    def add(self, test: Callable[[str], bool], allow: bool = True):
        self.add_condition(FilterCondition(test, allow))

    def filter(self, path: bytes) -> bool:
        allowed = False
        for condition in self._conditions:
            if condition.allow and allowed:
                continue  # Only a single allow condition needs to match
            matched = condition.evaluate(path)
            if matched:
                if condition.allow:
                    allowed = True
                else:
                    return False  # Any disallowed condition takes precedence
        return allowed


def matches_regex(regex: re.Pattern, string: bytes) -> bool:
    return regex.search(string) is not None


def filter_any(path: bytes) -> bool:
    return True


PATTERN_PHP = re.compile(
        br'\.(?:php(?:\d+)?|phtml)(\.|$)',
        re.IGNORECASE
    )
PATTERN_HTML = re.compile(
        br'\.(?:html?)(\.|$)',
        re.IGNORECASE
    )
PATTERN_JS = re.compile(
        br'\.(?:js|svg)(\.|$)',
        re.IGNORECASE
    )
PATTERN_IMAGES = re.compile(
        (
            br'\.(?:jpg|jpeg|mp3|avi|m4v|mov|mp4|gif|png|tiff?|svg|sql|js|tbz2?'  # noqa: E501
            br'|bz2?|xz|zip|tgz|gz|tar|log|err\d+)(\.|$)'
        ),
        re.IGNORECASE
    )


def filter_php(path: bytes) -> bool:
    return matches_regex(PATTERN_PHP, path)


def filter_html(path: bytes) -> bool:
    return matches_regex(PATTERN_HTML, path)


def filter_js(path: bytes) -> bool:
    return matches_regex(PATTERN_JS, path)


def filter_images(path: bytes) -> bool:
    return matches_regex(PATTERN_IMAGES, path)


class FilenameFilter:

    def __init__(self, value: bytes):
        self.value = value

    def __call__(self, path: bytes):
        filename = os.path.basename(path)
        return filename == self.value


class Filter:

    def __init__(self, pattern: Pattern[AnyStr]):
        self.pattern = pattern

    def __call__(self, path: bytes) -> bool:
        return matches_regex(self.pattern, path)


class InvalidPatternException(Exception):

    def __init__(self, pattern: bytes):
        self.pattern = pattern


def filter_pattern(regex: bytes) -> Callable[[bytes], bool]:
    try:
        pattern = re.compile(regex)
        return Filter(pattern)
    except re.error:
        raise InvalidPatternException(regex)



import time

from enum import Enum, auto
from typing import Dict, Optional, TextIO

from ..logging import log


class ProfileTimestamp:

    def __init__(self):
        self.clock_time = time.monotonic_ns()
        self.cpu_time = time.process_time_ns()


class TimeType(Enum):
    CLOCK = auto()
    CPU = auto()


class Timer:

    def __init__(self, time_type: TimeType, start: bool = True):
        self.time_type = time_type
        if start:
            self.start()
        else:
            self._start = None
        self._end = None

    def start(self) -> None:
        self._start = self._get_timestamp()

    def stop(self) -> None:
        self._end = self._get_timestamp()

    def _get_timestamp(self) -> int:
        raise NotImplementedError()

    def get_duration(self) -> int:
        if self._end is None or self._start is None or self._end < self._start:
            return 0
        return self._end - self._start


class ClockTimer(Timer):

    def __init__(self, start: bool = True):
        super().__init__(TimeType.CLOCK, start)

    def _get_timestamp(self) -> int:
        return time.monotonic_ns()


class CpuTimer(Timer):

    def __init__(self, start: bool = True):
        super().__init__(TimeType.CPU, start)

    def _get_timestamp(self) -> int:
        return time.process_time_ns()


def format_duration(duration: int) -> str:
    seconds = duration / 1000000000
    return str(round(seconds, 4)) + 's'


class ProfileEvent:

    def __init__(
                self,
                name: str,
                times: Dict[TimeType, int],
                is_global: bool = False
            ):
        self.name = name
        self.times = times
        self.is_global = is_global

    def get_time(self, time_type: TimeType = TimeType.CLOCK) -> Optional[int]:
        try:
            return self.times[time_type]
        except KeyError:
            return None

    def __str__(self) -> str:
        return ', '.join([
                type.name + ': ' + format_duration(time) for (type, time)
                in self.times.items()
            ])


def _get_times(*timers) -> Dict[TimeType, int]:
    times = {}
    for timer in timers:
        timer.stop()
        times[timer.time_type] = timer.get_duration()
    return times


class EventTimer:

    def __init__(self, name: str, start: bool = True, is_global: bool = False):
        self.name = name
        if start:
            self.start()
        else:
            self.clock_timer = None
            self.cpu_timer = None
        self.is_global = is_global

    def start(self) -> None:
        self.clock_timer = ClockTimer()
        self.cpu_timer = CpuTimer()

    def stop(self) -> Optional[ProfileEvent]:
        if self.clock_timer is None:
            return None
        return ProfileEvent(
                self.name,
                _get_times(self.clock_timer, self.cpu_timer),
                self.is_global
            )


class TimeAggregate:

    def __init__(self):
        self.average = 0
        self.min = None
        self.max = None
        self.total = 0
        self.count = 0

    def add(self, value: int) -> None:
        previous_count = self.count
        self.count += 1
        self.average = (
                (self.average * previous_count) + value
            ) / self.count
        self.min = value if self.min is None else min(self.min, value)
        self.max = value if self.max is None else max(self.max, value)
        self.total += value

    def __str__(self) -> str:
        average = format_duration(self.average)
        min = format_duration(self.min)
        max = format_duration(self.max)
        total = format_duration(self.total)
        return f'Total: {total}, Average: {average}, Min: {min}, Max: {max}'


class EventGroup:

    def __init__(self):
        self.events = []
        self.aggregates = {}

    def add(self, event: ProfileEvent) -> None:
        self.events.append(event)
        for time_type, duration in event.times.items():
            try:
                aggregate = self.aggregates[time_type]
            except KeyError:
                aggregate = TimeAggregate()
                self.aggregates[time_type] = aggregate
            aggregate.add(duration)


class ProfileWriter:

    def write(entry: str):
        raise NotImplementedError()


class LogProfileWriter(ProfileWriter):

    def __init__(self):
        self.write('=== PROFILING RESULTS ===')

    def write(self, entry: str):
        log.info(entry)


class FileProfileWriter(ProfileWriter):

    def __init__(self, file: TextIO):
        self.file = file

    def write(self, entry: str):
        self.file.write(entry)
        self.file.write('\n')


class ProfileWriterFactory:

    def __enter__(self) -> ProfileWriter:
        raise NotImplementedError()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        raise NotImplementedError()


class LogProfileWriterFactory(ProfileWriterFactory):

    def __enter__(self) -> ProfileWriter:
        return LogProfileWriter()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        return


class FileProfileWriterFactory(ProfileWriterFactory):

    def __init__(self, path: str):
        self.path = path
        self.file = None

    def __enter__(self) -> ProfileWriter:
        self.file = open(self.path, 'w')
        return FileProfileWriter(self.file)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.file.close()


class Profiler:

    def __init__(self):
        self.event_groups = {}
        self.global_events = EventGroup()
        self.timer = EventTimer('global', is_global=True)

    def complete(self) -> None:
        self.global_events.add(self.timer.stop())

    def add_event(self, event: ProfileEvent) -> None:
        try:
            group = self.event_groups[event.name]
        except KeyError:
            group = EventGroup()
            self.event_groups[event.name] = group
        group.add(event)
        if event.is_global:
            self.global_events.add(event)

    def _output_group(self, group: EventGroup, writer: ProfileWriter) -> None:
        for time_type, aggregate in group.aggregates.items():
            writer.write(f'\t{time_type.name}: {aggregate}')

    def output_results(self, writer: ProfileWriter) -> None:
        for name, group in self.event_groups.items():
            writer.write(f'{name}:')
            self._output_group(group, writer)
        writer.write('global:')
        self._output_group(self.global_events, writer)



from ..subcommands import Subcommand


class HelpSubcommand(Subcommand):

    def invoke(self) -> int:
        subcommand = None
        for argument in self.config.trailing_arguments:
            if subcommand is not None:
                raise Exception('Please specify a single subcommand')
            subcommand = argument
        self.helper.display_help(subcommand)
        return 0


factory = HelpSubcommand



import sys
import os
from collections import namedtuple
from configparser import ConfigParser, DuplicateSectionError
from multiprocessing import cpu_count
from typing import Optional, List, Dict, TextIO, Callable

from wordfence.util.input import prompt, prompt_yes_no, prompt_int, \
        InvalidInputException, InputException
from wordfence.util.io import ensure_directory_is_writable, \
        ensure_file_is_writable, resolve_path, IoException
from wordfence.api.licensing import License, LICENSE_URL
from wordfence.logging import log
from .config import load_config
from .context import CliContext
from .subcommands import SubcommandDefinition
from .licensing import LicenseManager, LicenseValidationFailure
from .terms_management import TERMS_URL, TermsManager
from .helper import Helper
from .mailing_lists import EMAIL_SIGNUP_MESSAGE


CONFIG_SECTION_DEFAULT = 'DEFAULT'
LEGACY_CONFIG_SECTION = 'SCAN'
LEGACY_CONFIG_KEYS = {
        'license',
        'cache_directory',
        'workers'
    }
LEGACY_CONVERSION_SECTION = 'MALWARE_SCAN'
MIN_WORKERS = 1


ConfigValue = namedtuple('ConfigValue', ['section', 'key', 'value'])


class ConfigFileManager:

    def __init__(
                self,
                config
            ):
        self.config = config
        self.parser = None
        self._read = False

    def initialize_parser(self) -> None:
        self.parser = ConfigParser()
        self._read = False

    def require_parser(self) -> None:
        if self.parser is None:
            self.initialize_parser()

    def require_section(self, section: str) -> None:
        self.require_parser()
        if section == CONFIG_SECTION_DEFAULT:
            return
        if self.parser.has_section(section):
            return
        try:
            self.parser.add_section(section)
        except DuplicateSectionError:
            pass

    def apply_update(self, update: ConfigValue) -> None:
        self.require_parser()
        self.require_section(update.section)
        self.parser.set(update.section, update.key, update.value)

    def resolve_ini_path(self) -> bytes:
        ini_path = self.config.ini_path if self.config.has_ini_file() \
            else self.config.configuration
        return resolve_path(ini_path)

    def read_existing_config(self, file: TextIO, ini_path: bytes) -> None:
        try:
            if not self._read:
                self.parser.read_file(file)
        except BaseException:  # noqa: B036
            log.warning(
                    'Failed to read existing config file at '
                    + os.fsdecode(ini_path) + '. existing data will be '
                    'truncated.'
                )
        self._read = True

    def write(self, updater: Callable[[], List[ConfigValue]]) -> None:
        # TODO: What if the INI file changes after the config is loaded?
        self.require_parser()
        ini_path = self.resolve_ini_path()
        ensure_file_is_writable(ini_path)
        open_mode = 'r' if self.config.has_ini_file() else 'w'
        with open(ini_path, open_mode + '+') as file:
            if self.config.has_ini_file():
                self.read_existing_config(file, ini_path)

            updates = updater()

            for update in updates:
                self.apply_update(update)

            file.truncate(0)
            file.seek(0)
            ini_path_str = os.fsdecode(ini_path)
            log.debug(f'Writing config to {ini_path_str}...')
            self.parser.write(file)
            self.written = True
            log.info(f'Config saved to {ini_path_str}')

    def read(self) -> List[ConfigValue]:
        values = []
        self.initialize_parser()
        ini_path = self.resolve_ini_path()
        try:
            with open(ini_path, 'r') as file:
                self.read_existing_config(file, ini_path)
            for section_name, section_proxy in self.parser.items():
                for key, value in section_proxy.items():
                    values.append(ConfigValue(section_name, key, value))
        except FileNotFoundError:
            log.debug(
                    'No existing config file found at '
                    + os.fsdecode(ini_path)
                )
        return values

    def delete_section(self, section: str) -> None:
        self.require_parser()
        self.parser.remove_section(section)


class Configurer:

    def __init__(
                self,
                context: CliContext,
                helper: Helper,
                license_manager: LicenseManager,
                terms_manager: TermsManager,
                subcommand_definitions: Dict[str, SubcommandDefinition],
                subcommand_definition: Optional[SubcommandDefinition] = None
            ):
        self.context = context
        self.config = context.config
        self.helper = helper
        self.all_config = {}
        self.all_config[context.config.subcommand] = context.config
        self.config_values = []
        self.license_manager = license_manager
        self.terms_manager = terms_manager
        self.subcommand_definition = subcommand_definition
        self.subcommand_definitions = subcommand_definitions
        self.overwrite = None
        self.request_license = None
        self.workers = None
        self.default = False
        self.written = False
        self.config_file_manager = None

    def get_config_file_manager(self) -> ConfigFileManager:
        if self.config_file_manager is None:
            self.config_file_manager = ConfigFileManager(self.config)
        return self.config_file_manager

    def get_config(self, subcommand: str):
        if subcommand not in self.all_config:
            self.all_config[subcommand], _subcommand_definition = load_config(
                        self.subcommand_definitions,
                        self.helper,
                        subcommand
                    )
        return self.all_config[subcommand]

    def supports_option(self, name: str) -> bool:
        if self.subcommand_definition is None:
            return False
        return self.subcommand_definition.accepts_option(name)

    def has_base_config(self) -> bool:
        if self.config.license is None:
            return False
        try:
            ensure_directory_is_writable(self.config.cache_directory)
        except IoException:
            log.warning(
                    f'Cache directory at {self.config.cache_directory} does'
                    'not appear to be writable. Please correct the permissions'
                    ' or specify an alternate path for the cache.'
                )
            return False
        return True

    def _prompt_overwrite(self) -> bool:
        if not self.overwrite and self.config.has_ini_file():
            overwrite = prompt_yes_no(
                    'An existing configuration file was found at '
                    + os.fsdecode(self.config.ini_path) +
                    ', do you want to update it?',
                    default=False
                )
            return overwrite
        return True

    def _prompt_for_license(self) -> License:

        if self.config.is_from_cli('license'):
            return self.license_manager.validate_license(self.config.license)

        if self.config.license is not None:
            print(f'Current license: {self.config.license}')
            change_license = self.request_license \
                or self.default \
                or prompt_yes_no(
                    'An existing license was found, '
                    'would you like to change it?',
                    default=False
                )
            if not change_license:
                try:
                    return self.license_manager.validate_license(
                            self.config.license
                        )
                except LicenseValidationFailure as failure:
                    print(failure.message)
                    print(
                            'Your existing license is invalid. Please specify '
                            'a valid license.'
                        )

        request_free = self.default or self.request_license or prompt_yes_no(
                'Would you like to automatically request a free Wordfence CLI'
                ' license?',
                default=True
            )
        if not request_free:
            print(f'Please visit {LICENSE_URL} to obtain a license key.')

        if request_free:
            terms_accepted = self.config.accept_terms or prompt_yes_no(
                    'Your access to and use of Wordfence CLI Free edition is '
                    'subject to the Wordfence CLI License Terms and '
                    f'Conditions set forth at {TERMS_URL}. By entering "y" '
                    'and selecting Enter, you agree that you have read and '
                    'accept the Wordfence CLI License Terms and Conditions.',
                    default=False
                )
            if terms_accepted:
                license = self.license_manager.request_free_license(
                        terms_accepted
                    )
                self.terms_manager.record_acceptance(
                        license=license,
                        remote=False
                    )
                print(
                        'Free Wordfence CLI license obtained successfully: '
                        f'{license}'
                    )
                return license
            else:
                print(
                        'A license cannot be obtained automatically without'
                        ' agreeing to the Wordfence CLI License Terms and '
                        'Conditions.'
                    )

        license = prompt(
                'License',
                self.config.license,
                transformer=self.license_manager.validate_license
            )
        return license

    def _prompt_for_cache_directory(self) -> str:

        def _validate_writable(directory: str) -> None:
            try:
                ensure_directory_is_writable(directory)
            except IoException as e:
                raise InvalidInputException(
                        f'Directory {directory} is not writable'
                    ) from e
            return os.fsencode(directory)

        if self.config.is_from_cli('cache_directory') or self.default:
            _validate_writable(self.config.cache_directory)
            return self.config.cache_directory

        directory = prompt(
                'Cache directory',
                os.fsdecode(self.config.cache_directory),
                transformer=_validate_writable
            )
        return directory

    def _prompt_for_worker_count(self) -> int:
        if self.workers is not None:
            return self.workers
        if self.default:
            return MIN_WORKERS
        cpus = cpu_count()
        config = self.get_config('malware-scan')
        workers = max(int(config.workers), MIN_WORKERS)
        processes = prompt_int(
                    f'Number of worker processes ({cpus} CPUs available)',
                    workers,
                    min=MIN_WORKERS
                )
        return processes

    def read_config(self) -> List[ConfigValue]:
        manager = self.get_config_file_manager()
        return manager.read()

    def update_config(
                self,
                key: str,
                value: str,
                section: str = 'DEFAULT'
            ) -> None:
        self.config_values.append(
                ConfigValue(section, key, str(value))
            )
        if self.supports_option(key):
            setattr(self.config, key, value)

    def prompt_for_all(self) -> List[ConfigValue]:
        cache_directory = self._prompt_for_cache_directory()
        self.update_config(
                'cache_directory',
                os.fsdecode(cache_directory)
            )
        self.context.set_up_cache(cache_directory)
        self.license = self._prompt_for_license()
        self.update_config(
                'license',
                self.license.key
            )
        self.update_config(
                'workers',
                self._prompt_for_worker_count(),
                'MALWARE_SCAN'
            )
        return self.config_values

    def prompt_for_config(self, overwrite: bool = False) -> bool:
        try:
            if not overwrite and not self._prompt_overwrite():
                return False
            manager = self.get_config_file_manager()
            has_existing_config = self.config.has_ini_file()
            manager.write(updater=self.prompt_for_all)
            self.license_manager.set_license(self.license)
            if has_existing_config:
                log.info(
                        "The configuration for Wordfence CLI has been "
                        "successfully updated."
                    )
            else:
                log.info(
                        "Wordfence CLI has been successfully configured and "
                        "is now ready for use."
                    )
            log.info(EMAIL_SIGNUP_MESSAGE)
            return True
        except InputException:
            self._handle_prompt_error()

    def convert_legacy_config(self) -> bool:
        values = self.read_config()
        has_legacy_config = False
        for value in values:
            if not value.section == LEGACY_CONFIG_SECTION:
                continue
            if value.key in LEGACY_CONFIG_KEYS:
                setattr(self.config, value.key, value.value)
                has_legacy_config = True
            else:
                self.update_config(
                        value.key,
                        value.value,
                        LEGACY_CONVERSION_SECTION
                    )
        if not has_legacy_config:
            return False
        should_convert = prompt_yes_no(
                'A configuration file for an older version of Wordfence CLI '
                'was detected; would you like to update it now?',
                default=True
            )
        if should_convert:
            self.config_file_manager.delete_section(LEGACY_CONFIG_SECTION)
            self.prompt_for_config(overwrite=True)
        return True

    def prompt_for_missing_config(self) -> bool:
        try:
            should_configure = prompt_yes_no(
                    'Wordfence CLI cannot be used until it has been '
                    'configured. Would you like to configure it now?',
                    default=False
                )
            if should_configure:
                self.prompt_for_config()
                return True
            else:
                return False
        except InputException:
            self._handle_prompt_error()

    def _handle_prompt_error(self) -> None:
        print(
                'Wordfence CLI does not appear to be running interactively '
                'and cannot prompt for configuration. Please run Wordfence '
                'CLI in a terminal, specify the configuration options using '
                'the various command line parameters, or set up a '
                'configuration file manually. Run wordfence configure --help '
                'for additional information.'
            )
        sys.exit(1)

    def check_config(self) -> bool:
        if self.has_base_config():
            return True
        else:
            if not self.convert_legacy_config():
                self.prompt_for_missing_config()
            return False



from typing import Optional, Any, Set


class ValidationException(Exception):

    def __init__(self, key: list, message: str, value=None):
        self.key = key
        self.value = value
        super().__init__(
                self.get_key_as_string() +
                ': ' +
                message +
                ', received: ' +
                repr(value)
            )

    def get_key_as_string(self) -> str:
        return '.'.join([str(component) for component in self.key])


class Validator:

    def validate(self, data, parent_key: Optional[list] = None):
        pass

    def validate_type(self, key, value, expected_type) -> None:
        if isinstance(expected_type, Validator):
            expected_type.validate(value, key)
        elif not isinstance(value, expected_type):
            raise ValidationException(
                    key,
                    'Value must be of type ' + str(expected_type),
                    value
                )


class DictionaryValidator(Validator):

    def __init__(
                self,
                expected: Optional[dict] = None,
                validator: Validator = None,
                allow_empty: bool = False,
                optional_keys: Optional[Set[str]] = None
            ):
        self.expected = expected if expected is not None else dict()
        self.validator = validator
        self.allow_empty = allow_empty
        self.optional_keys = optional_keys if optional_keys is not None else {}

    def _validate_expected_fields(self, data: dict, parent_key: list) -> None:
        for key, expected_type in self.expected.items():
            aggregate_key = parent_key + [key]
            try:
                value = data[key]
                self.validate_type(aggregate_key, value, expected_type)
            except KeyError:
                if key not in self.optional_keys:
                    raise ValidationException(aggregate_key, 'Key not present')

    def _validate_all_fields(self, data: dict, parent_key: list) -> None:
        if self.validator is None:
            return
        for key, value in data.items():
            if key not in self.expected:
                self.validator.validate(value, parent_key + [key])

    def validate(self, data, parent_key: Optional[list] = None) -> None:
        if parent_key is None:
            parent_key = []
        if not isinstance(data, dict):
            raise ValidationException(
                    parent_key,
                    'Element must be a dictionary',
                    data
                )
        if self.allow_empty and len(data) == 0:
            return
        self._validate_expected_fields(data, parent_key)
        self._validate_all_fields(data, parent_key)

    def add_field(self, key: Any, expected):
        self.expected[key] = expected


class ListValidator(Validator):

    def __init__(self, expected):
        self.expected = expected

    def validate(self, data, parent_key: Optional[list] = None) -> None:
        if parent_key is None:
            parent_key = []
        if not isinstance(data, list):
            raise ValidationException(
                    parent_key,
                    'Element must be a list',
                    data
                )
        if isinstance(self.expected, dict):
            for index, expected_type in self.expected.items():
                key = parent_key + [index]
                try:
                    value = data[index]
                    self.validate_type(key, data[index], expected_type)
                except IndexError:
                    raise ValidationException(
                            key,
                            'Index does not exist in list'
                        )
        else:
            for index, value in enumerate(data):
                self.validate_type(parent_key + [index], value, self.expected)


class AllowedValueValidator(Validator):

    def __init__(self, allowed: set):
        self.allowed = allowed

    def validate(self, data, parent_key: Optional[list] = None) -> None:
        for value in self.allowed:
            if data == value:
                return
        raise ValidationException(
                parent_key,
                'Value is not in allowed set: ' + repr(data)
            )


class OptionalValueValidator(Validator):

    def __init__(self, expected):
        self.expected = expected

    def validate(self, data, parent_key: Optional[list] = None) -> None:
        if data is None:
            return
        if parent_key is None:
            parent_key = []
        if isinstance(self.expected, Validator):
            self.expected.validate(data, parent_key)
        else:
            self.validate_type(parent_key, data, self.expected)


class NumberValidator(Validator):

    def __init__(self):
        pass

    def validate(self, data, parent_key: Optional[list] = None) -> None:
        if isinstance(data, int) or isinstance(data, float):
            return
        raise ValidationException(
                parent_key,
                'Value is not a valid number: ' + repr(data)
            )



from enum import Enum


class MailingList(Enum):

    TERMS = (
            'https://www.wordfence.com/products/wordfence-cli/#terms'
        )
    WORDPRESS_SECURITY = (
            'https://www.wordfence.com/subscribe-to-the-wordfence-email-list/'
        )

    def __init__(self, registration_url):
        self.registration_url = registration_url


EMAIL_SIGNUP_MESSAGE = (
    "Register to receive updated Wordfence CLI Terms of Service via email "
    f"at {MailingList.TERMS.registration_url}. Join our WordPress Security "
    f"mailing list at {MailingList.WORDPRESS_SECURITY.registration_url} "
    "to get security alerts, news, and research directly to your inbox."
)



import os
import mmap
import math

from typing import Optional


class DirectIoBuffer:

    def __init__(self, max_chunk_size: int = mmap.PAGESIZE):
        self.max_chunk_size = max_chunk_size
        self.buffer_size = (
                math.ceil(max_chunk_size / mmap.PAGESIZE) * mmap.PAGESIZE
            )
        self.buffer = mmap.mmap(-1, self.buffer_size)
        self.buffer_view = memoryview(self.buffer)
        self.buffers = [self.buffer]

    def seek(self, position: int = 0) -> list:
        self.buffer.seek(position)

    def read(self, length: int) -> bytes:
        return self.buffer.read(length)


class DirectIoReader:

    def __init__(self, path: str, buffer: DirectIoBuffer):
        self.fd = os.open(path, os.O_RDONLY | os.O_DIRECT)
        self.buffer = buffer
        self.offset = 0

    def read(self, limit: Optional[int] = None) -> bytes:
        read_offset = math.floor(self.offset / mmap.PAGESIZE) * mmap.PAGESIZE
        skip = self.offset % mmap.PAGESIZE
        read_length = os.preadv(self.fd, self.buffer.buffers, read_offset)
        read_length -= skip
        read_length = min(read_length, limit)
        self.offset += read_length
        self.buffer.seek(skip)
        return self.buffer.read(read_length)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        os.close(self.fd)



from typing import Union, Optional

from .exceptions import ApiException


LICENSE_URL = 'https://www.wordfence.com/products/wordfence-cli/'


class License:

    def __init__(self, key: str):
        self.key = key
        self.paid = False

    def __eq__(self, other):
        return other.key == self.key

    def __str__(self) -> str:
        return self.key


def to_license(license: Union[License, str]) -> License:
    if isinstance(license, License):
        return license
    return License(license)


class LicenseRequiredException(ApiException):

    def __init__(self):
        super().__init__(
                'License required',
                'A valid Wordfence CLI license is required'
            )


class LicenseSpecific:

    def __init__(self, license: Optional[License]):
        self.license = license

    def is_compatible_with_license(self, license: License):
        return self.license is None or self.license == license

    def assign_license(self, license: Optional[License]):
        self.license = license

    def clear_license(self):
        self.assign_license(None)



import sys

from typing import Optional, Callable, Any


class InputException(Exception):
    pass


class InvalidInputException(InputException):

    def __init__(self, message: str):
        self.message = message


class NoInputException(InputException):
    pass


class NoTerminalException(InputException):
    pass


def has_terminal_output() -> bool:
    return sys.stdout is not None \
            and sys.stdout.isatty()


def has_terminal_input() -> bool:
    return sys.stdin is not None \
            and sys.stdin.isatty()


def has_terminal() -> bool:
    return has_terminal_input() and has_terminal_output()


def prompt(
            message: str,
            default: Optional[str] = None,
            transformer: Optional[Callable[[str], Any]] = None,
            allow_empty: bool = False
        ) -> Any:
    if not has_terminal():
        raise NoTerminalException('Interactive prompts require a terminal')
    default_message = ''
    if default is not None:
        default_message = f' (default: {default})'
    while True:
        try:
            response = input(f'{message}{default_message}: ')
        except EOFError as error:
            raise NoInputException('Unable to read input') from error
        if len(response) == 0 and not allow_empty:
            response = default
        if transformer is not None:
            try:
                return transformer(response)
            except InvalidInputException as e:
                print(e.message)
        else:
            return response


def transform_yn_to_bool(response: str) -> Any:
    lower_response = response.lower()
    if lower_response == 'y':
        return True
    elif lower_response == 'n':
        return False
    else:
        raise InvalidInputException(
                f'Invalid response: "{response}", please enter "y" or "n"'
            )


def transform_str_to_int(response: str) -> int:
    try:
        if response.isascii() and response.isdigit():
            return int(response)
    except ValueError:
        pass
    raise InvalidInputException(
            'Please enter a valid integer'
        )


def initialize_str_to_int_transformer(
            min: Optional[int] = None,
            max: Optional[int] = None
        ):
    def transformer(response: str) -> int:
        value = transform_str_to_int(response)
        if min is not None and value < min:
            raise InvalidInputException(
                    f'Please enter a value that is at least {min}'
                )
        if max is not None and value > max:
            raise InvalidInputException(
                    f'Please enter a value that is no greater than {max}'
                )
        return value
    return transformer


def prompt_yes_no(message: str, default: Optional[bool] = None) -> bool:
    default_string = None
    if default is not None:
        default_string = 'y' if default else 'n'
    return prompt(
            message=f'{message} [y/n]',
            default=default_string,
            transformer=transform_yn_to_bool
        )


def prompt_int(
            message: str,
            default: Optional[int] = None,
            min: Optional[int] = None,
            max: Optional[int] = None
        ) -> int:
    default_string = None
    if default is not None:
        default_string = str(default)
    transformer = initialize_str_to_int_transformer(min, max)
    return prompt(
            message,
            default=default_string,
            transformer=transformer
        )



import sys
from enum import IntEnum


def supports_colors() -> bool:
    # TODO: Implement more detailed checks for color support
    return sys.stdout is not None and sys.stdout.isatty()


class Color(IntEnum):
    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    RESET = 0


ESC = '\x1b'


def escape(color: Color, bold: bool = False) -> str:
    fields = []
    if bold:
        fields.append('1')
    else:
        fields.append('22')
    fields.append(str(color.value))
    sequence = ';'.join(fields)
    return f'{ESC}[{sequence}m'


RESET = escape(color=Color.RESET)



import os
import queue
import time
import traceback
import dataclasses
from ctypes import c_bool, c_uint, c_char
from enum import IntEnum
from multiprocessing import Queue, Process, Value, Array, get_start_method
from dataclasses import dataclass
from typing import Set, Optional, Callable, Dict, NamedTuple, Tuple, List, \
    Union, BinaryIO
from logging import Handler

from .exceptions import ScanningException, ScanningIoException
from .matching import MatchEngine, MatchEngineOptions, Matcher, MatchWorkspace
from .filtering import FileFilter, filter_any
from ..util import timing
from ..util.io import StreamReader, is_symlink_loop, is_symlink_and_loop, \
    get_all_parents, PathSet
from ..util.direct_io import DirectIoBuffer, DirectIoReader
from ..util.units import scale_byte_unit
from ..logging import log, remove_initial_handler, VERBOSE
from ..util.profiling import Profiler, ProfileEvent, EventTimer, \
    LogProfileWriterFactory, FileProfileWriterFactory

MAX_PENDING_FILES = 1000  # Arbitrary limit
MAX_PENDING_RESULTS = 100
QUEUE_READ_TIMEOUT = 0
RESULT_QUEUE_READ_TIMEOUT = 1
PROGRESS_UPDATE_INTERVAL = 100
DEFAULT_CHUNK_SIZE = 1024 * 1024
PATH_NAME_LIMIT = 4096
FILE_LOCATOR_WORKER_INDEX = 0
"""Used by the file locator process when sending events"""


USES_FORK = get_start_method() == 'fork'


class ExceptionContainer(Exception):

    def __init__(self, exception: BaseException, trace: str = None):
        self.exception = exception
        if trace is None:
            self.trace = traceback.format_exc()
        else:
            self.trace = trace
        message = str(self.exception)
        super().__init__(
                f'An exception occurred in a child process: {message}'
            )

    def __reduce__(self) -> Tuple:
        return (
                self.__class__,
                (
                    self.exception,
                    self.trace
                )
            )


class ScanConfigurationException(ScanningException):
    pass


class ScanEventType(IntEnum):
    COMPLETED = 0
    FILE_QUEUE_EMPTIED = 1
    FILE_PROCESSED = 2
    EXCEPTION = 3
    FATAL_EXCEPTION = 4
    PROGRESS_UPDATE = 5
    LOG_MESSAGE = 6
    PROFILE_EVENT = 7


class ScanEvent:

    # TODO: Define custom (more compact) pickle serialization format for this
    # class as a potential performance improvement

    def __init__(
                self,
                type: int,
                data=None,
                worker_index: Optional[int] = None
            ):
        self.type = type
        self.data = data
        self.worker_index = worker_index


class ScanProfileEvent(ScanEvent):

    def __init__(
                self,
                event: ProfileEvent,
                worker_index: Optional[int] = None
            ):
        super().__init__(ScanEventType.PROFILE_EVENT, event, worker_index)


def _put_profile_event(
            queue: Queue,
            event: Optional[Union[ProfileEvent, EventTimer]]
        ) -> None:
    if event is None:
        return
    if isinstance(event, EventTimer):
        event = event.stop()
    queue.put(ScanProfileEvent(event))


def _event_timer(
            condition: bool,
            name: str,
            is_global: bool = False
        ) -> Optional[EventTimer]:
    if not condition:
        return None
    return EventTimer(name, is_global=is_global)


class EventQueueLogHandler(Handler):

    def __init__(self, event_queue: Queue, worker_index: int):
        self._event_queue = event_queue
        self._worker_index = worker_index
        Handler.__init__(self)

    def emit(self, record):
        data = {
                'level': record.levelno,
                'message': record.getMessage()
            }
        self._event_queue.put(
            ScanEvent(
                ScanEventType.LOG_MESSAGE,
                data,
                worker_index=self._worker_index
            )
        )


def use_event_queue_log_handler(event_queue: Queue, worker_index: int) -> None:
    handler = EventQueueLogHandler(
            event_queue,
            worker_index
        )
    remove_initial_handler()
    log.addHandler(handler)


@dataclass
class Options:
    paths: Set[bytes]
    match_engine_options: MatchEngineOptions
    workers: int = 1
    chunk_size: int = DEFAULT_CHUNK_SIZE
    path_source: Optional[StreamReader] = None
    scanned_content_limit: Optional[int] = None
    file_filter: Optional[FileFilter] = None
    allow_io_errors: bool = False
    debug: bool = False
    logging_initializer: Callable[[], None] = None
    match_engine: MatchEngine = MatchEngine.get_default()
    profile: bool = False,
    profile_path: Optional[str] = None,
    direct_io: bool = False


class Status(IntEnum):
    LOCATING_FILES = 0
    PROCESSING_FILES = 1
    COMPLETE = 2
    FAILED = 3


class FileLocator:

    def __init__(
                self,
                path: bytes,
                queue: Queue,
                file_filter: FileFilter,
                allow_io_errors: bool = False,
                scanned_paths: Optional[PathSet] = None
            ):
        self.path = path
        self.queue = queue
        self.file_filter = file_filter
        self.allow_io_errors = allow_io_errors
        self.scanned_paths = scanned_paths if scanned_paths is not None \
            else PathSet()
        self.located_count = 0
        self.skipped_count = 0

    def _is_loop(
                self,
                path: bytes,
                parents: Optional[List[bytes]] = None
            ) -> bool:
        if is_symlink_loop(path, parents):
            log.warning('Symlink loop detected at ' + os.fsdecode(path))
            return True
        return False

    def _handle_io_error(self, error: OSError, path: bytes) -> None:
        detail = str(error)
        if self.allow_io_errors:
            log.warning(
                    'Path ' + os.fsdecode(path) + ' could not be scanned due '
                    f'to an IO error and will be skipped. ({detail})'
                )
        else:
            raise ScanningIoException(
                    f'Directory search of {path} failed ({detail})'
                ) from error

    def search_directory(self, path: bytes, parents: Optional[list] = None):
        try:
            if parents is None:
                parents = get_all_parents(path)
                self.scanned_paths.add(path)
            contents = os.scandir(path)
        except OSError as os_error:
            self._handle_io_error(os_error, path)
            return
        for item in contents:
            item_path = item.path
            try:
                if item.is_symlink():
                    item_path = os.path.realpath(item.path)
                    if item_path in self.scanned_paths:
                        continue
                    # This intentionally uses the unresolved path
                    if self._is_loop(item.path, parents):
                        continue
                if item.is_dir():
                    self.scanned_paths.add(item_path)
                    yield from self.search_directory(
                            item_path,
                            parents + get_all_parents(item_path)
                        )
                elif item.is_file():
                    if not self.file_filter.filter(item_path):
                        self.skipped_count += 1
                        continue
                    self.located_count += 1
                    yield item_path
            except OSError as os_error:
                self._handle_io_error(os_error, item_path)

    def _push_file(self, path: bytes) -> None:
        if path in self.scanned_paths:
            log.warning(
                    'Skipping already queued path: ' + os.fsdecode(path)
                )
        else:
            log.log(VERBOSE, 'File added to scan queue: ' + os.fsdecode(path))
            self.queue.put(path)
            self.scanned_paths.add(path)

    def locate(self):
        real_path = os.path.realpath(self.path)
        if os.path.isdir(real_path):
            for path in self.search_directory(real_path):
                self._push_file(path)
        else:
            if not is_symlink_and_loop(self.path):
                self._push_file(real_path)


class FileLocatorProcess(Process):

    def __init__(
                self,
                input_queue_size: int = 10,
                output_queue_size: int = MAX_PENDING_FILES,
                file_filter: FileFilter = None,
                use_log_events: bool = False,
                event_queue: Optional[Queue] = None,
                allow_io_errors: bool = False,
                logging_initializer: Callable[[], None] = None,
                profile: bool = False
            ):
        self._input_queue = Queue(input_queue_size)
        self.output_queue = Queue(output_queue_size)
        self.file_filter = file_filter \
            if file_filter is not None \
            else FileFilter([filter_any])
        if use_log_events and not event_queue:
            raise ValueError('Using log events requires an event queue')
        self._use_log_events = use_log_events
        self._event_queue = event_queue
        self.allow_io_errors = allow_io_errors
        self._logging_initializer = logging_initializer
        self.profile = profile
        self._path_count = 0
        self._skipped_count = Value('i', 0)
        super().__init__(name='file-locator')

    def add_path(self, path: bytes) -> bool:
        try:
            self._input_queue.put(path, block=False)
            self._path_count += 1
            log.info('Scanning path: ' + os.fsdecode(path))
            return True
        except queue.Full:
            return False

    def finalize_paths(self) -> bool:
        try:
            self._input_queue.put(None, block=False)
        except queue.Full:
            return False
        if self._path_count < 1:
            raise ScanConfigurationException(
                    'At least one scan path must be specified'
                )
        return True

    def get_next_file(self):
        return self.output_queue.get()

    def _put_profile_event(
                self,
                event: Optional[Union[ProfileEvent, EventTimer]]
            ) -> None:
        if not self.profile:
            return
        _put_profile_event(self._event_queue, event)

    def run(self):
        timer = _event_timer(self.profile, 'file_locator_all', is_global=True)
        if self._logging_initializer is not None:
            self._logging_initializer()
        if self._use_log_events:
            use_event_queue_log_handler(
                    self._event_queue,
                    FILE_LOCATOR_WORKER_INDEX
                )
        try:
            skipped_count = 0
            scanned_paths = PathSet()
            while (path := self._input_queue.get()) is not None:
                path_timer = _event_timer(self.profile, 'file_locator_path')
                locator = FileLocator(
                        path=path,
                        file_filter=self.file_filter,
                        queue=self.output_queue,
                        allow_io_errors=self.allow_io_errors,
                        scanned_paths=scanned_paths
                    )
                locator.locate()
                skipped_count += locator.skipped_count
                self._put_profile_event(path_timer)
        except ScanningException as exception:
            self.output_queue.put(ExceptionContainer(exception))
        self._skipped_count.value = skipped_count
        self.output_queue.put(None)
        self._put_profile_event(timer)

    def get_skipped_count(self) -> int:
        return self._skipped_count.value


class ScanProgressMonitor(Process):

    def __init__(self, status: Value, event_queue: Queue):
        super().__init__(name='progress-monitor')
        self._event_queue = event_queue
        self._status = status

    def is_scan_running(self) -> bool:
        status = self._status.value
        return status != Status.COMPLETE and status != Status.FAILED

    def run(self):
        interval = PROGRESS_UPDATE_INTERVAL / 1000  # MS to seconds
        while self.is_scan_running():
            time.sleep(interval)
            self._event_queue.put(ScanEvent(ScanEventType.PROGRESS_UPDATE))


class ScanWorker(Process):

    def __init__(
                self,
                index: int,
                status: Value,
                work_queue: Queue,
                event_queue: Queue,
                matcher: Matcher,
                chunk_size: int = DEFAULT_CHUNK_SIZE,
                scanned_content_limit: Optional[int] = None,
                use_log_events: bool = False,
                allow_io_errors: bool = False,
                logging_initializer: Callable[[], None] = None,
                profile: bool = False,
                direct_io: bool = False
            ):
        self.index = index
        self._status = status
        self._work_queue = work_queue
        self._event_queue = event_queue
        self._matcher = matcher
        self._chunk_size = chunk_size
        self._working = True
        self._scanned_content_limit = scanned_content_limit
        self._use_log_events = use_log_events
        self._allow_io_errors = allow_io_errors
        self._logging_initializer = logging_initializer
        self._profile = profile
        self._opener = self._open_direct if direct_io else self._open
        self._direct_io = direct_io
        self.complete = Value(c_bool, lock=False)
        self.last_file = Array(c_char, PATH_NAME_LIMIT + 1, lock=False)
        self._timer = None
        super().__init__(name=self._generate_name())

    def _generate_name(self) -> str:
        return 'worker-' + str(self.index)

    def _open(self, path: str) -> BinaryIO:
        return open(path, 'rb')

    def _open_direct(self, path: str) -> DirectIoReader:
        return DirectIoReader(path, self._direct_io_buffer)

    def work(self):
        self._timer = _event_timer(
                self._profile,
                'scan_worker',
                is_global=True
            )
        if self._direct_io:
            self._direct_io_buffer = DirectIoBuffer(self._chunk_size)
        try:
            self._working = True
            self._matcher.prepare(thread=True)
            log.debug(f'Worker {self.index} started, PID:' + str(os.getpid()))
            with self._matcher.create_workspace() as workspace:
                while self._working:
                    try:
                        item = self._work_queue.get(timeout=QUEUE_READ_TIMEOUT)
                        if item is None:
                            self._put_event(ScanEventType.FILE_QUEUE_EMPTIED)
                            self._complete()
                        elif isinstance(item, ExceptionContainer):
                            if isinstance(item.exception, ScanningIoException):
                                self._put_io_error(item)
                            else:
                                self._put_event(
                                        ScanEventType.FATAL_EXCEPTION,
                                        {'exception': item}
                                    )
                        else:
                            timer = _event_timer(self._profile, 'process_file')
                            try:
                                self._process_file(item, workspace)
                            except OSError as error:
                                self._put_io_error(ExceptionContainer(error))
                            except Exception as error:
                                self._put_error(ExceptionContainer(error))
                            self._put_profile_event(timer)
                    except queue.Empty:
                        if self._status.value == Status.PROCESSING_FILES:
                            self._complete()
        except Exception as error:
            self._put_error(ExceptionContainer(error))

    def _put_event(self, event_type: ScanEventType, data: dict = None) -> None:
        if data is None:
            data = {}
        self._event_queue.put(
                ScanEvent(event_type, data, worker_index=self.index),
            )

    def _put_error(self, error, fatal: bool = True) -> None:
        event_type = ScanEventType.FATAL_EXCEPTION if fatal \
                else ScanEventType.EXCEPTION
        self._put_event(
                event_type,
                {'exception': error}
            )

    def _put_io_error(self, error) -> None:
        self._put_error(error, not self._allow_io_errors)

    def _put_profile_event(
                self,
                event: Optional[Union[ProfileEvent, EventTimer]]
            ) -> None:
        _put_profile_event(self._event_queue, event)

    def _complete(self):
        self._working = False
        self.complete.value = True
        if self._timer is not None:
            self._put_profile_event(self._timer.stop())
        self._put_event(ScanEventType.COMPLETED)

    def is_complete(self) -> bool:
        return self.complete.value

    def _get_next_chunk_size(self, length: int) -> int:
        if self._scanned_content_limit is None:
            return self._chunk_size
        elif length >= self._scanned_content_limit:
            return 0
        else:
            return min(self._scanned_content_limit - length, self._chunk_size)

    def _process_file(self, path: str, workspace: Optional[MatchWorkspace]):
        log.log(VERBOSE, 'Processing file: ' + os.fsdecode(path))
        self.last_file.value = path[:PATH_NAME_LIMIT] + b'\0'
        open_timer = _event_timer(self._profile, 'open_file')
        with self._opener(path) as file, \
                self._matcher.create_context() as context:
            self._put_profile_event(open_timer)
            length = 0
            while (chunk_size := self._get_next_chunk_size(length)):
                chunk_timer = _event_timer(self._profile, 'read_chunk')
                chunk = file.read(chunk_size)
                self._put_profile_event(chunk_timer)
                if not chunk:
                    break
                first = length == 0
                length += len(chunk)
                match_timer = _event_timer(self._profile, 'match_chunk')
                try:
                    if context.process_chunk(
                                chunk,
                                start=first,
                                workspace=workspace
                            ):
                        break
                finally:
                    self._put_profile_event(match_timer)
            context.finalize_content()
            self._put_event(
                    ScanEventType.FILE_PROCESSED,
                    {
                        'path': path,
                        'length': length,
                        'matches': context.matches,
                        'timeouts': context.timeouts
                    }
                )

    def run(self):
        if self._logging_initializer is not None:
            self._logging_initializer()
        if self._use_log_events:
            use_event_queue_log_handler(self._event_queue, self.index)
        self.work()


class ScanResult:

    def __init__(
                self,
                path: bytes,
                read_length: int,
                matches: Dict[int, bytes],
                timeouts: Set[int],
                timestamp: float = None
            ):
        self.path = path
        self.read_length = read_length
        self.matches = matches
        self.timeouts = timeouts
        self.timestamp = timestamp if timestamp is not None else time.time()

    def has_matches(self) -> bool:
        return len(self.matches) > 0

    def get_timeout_count(self) -> int:
        return len(self.timeouts)


class ScanMetrics:

    def __init__(self, worker_count: int):
        self.counts = self._initialize_int_metric(worker_count)
        self.bytes = self._initialize_int_metric(worker_count)
        self.matches = self._initialize_int_metric(worker_count)
        self.timeouts = self._initialize_int_metric(worker_count)
        self.skipped_files = 0
        self.failed_files = 0

    def _initialize_int_metric(self, worker_count: int):
        return [0] * worker_count

    def record_result(self, worker_index: int, result: ScanResult):
        self.counts[worker_index] += 1
        self.bytes[worker_index] += result.read_length
        if result.has_matches():
            self.matches[worker_index] += 1
        self.timeouts[worker_index] += result.get_timeout_count()

    def _aggregate_int_metric(self, metric: list) -> int:
        total = 0
        for value in metric:
            total += value
        return total

    def get_total_count(self) -> int:
        return self._aggregate_int_metric(self.counts)

    def get_total_bytes(self) -> int:
        return self._aggregate_int_metric(self.bytes)

    def get_total_matches(self) -> int:
        return self._aggregate_int_metric(self.matches)

    def get_total_timeouts(self) -> int:
        return self._aggregate_int_metric(self.timeouts)

    def get_int_metric(self, metric: str, worker_index: Optional[int] = None):
        values = getattr(self, metric)
        if worker_index is not None:
            return values[worker_index]
        else:
            return self._aggregate_int_metric(values)


class ScanProgressUpdate:

    def __init__(self, elapsed_time: int, metrics: ScanMetrics):
        self.elapsed_time = elapsed_time
        self.metrics = metrics


ScanResultCallback = Callable[[ScanResult], None]
ProgressReceiverCallback = Callable[[ScanProgressUpdate], None]
ScanFinishedCallback = Callable[
        [ScanMetrics, timing.Timer, Optional[Profiler]], None
    ]


class ScanFinishedMessages(NamedTuple):
    results: str
    timeouts: Optional[str]
    skipped: Optional[str]
    failed: Optional[str]


def get_scan_finished_messages(
            metrics: ScanMetrics,
            timer: timing.Timer
        ) -> ScanFinishedMessages:
    match_count = metrics.get_total_matches()
    total_count = metrics.get_total_count()
    byte_value = scale_byte_unit(metrics.get_total_bytes())
    elapsed_time = round(timer.get_elapsed())
    timeout_count = metrics.get_total_timeouts()
    timeouts_message = None
    if timeout_count > 0:
        timeouts_message = f'{timeout_count} timeout(s) occurred during scan'
    results_message = (f'Found {match_count} suspicious file(s) after '
                       f'processing {total_count} file(s) containing '
                       f'{byte_value} over {elapsed_time} second(s)')

    if metrics.skipped_files > 0:
        skipped_message = (
                f'{metrics.skipped_files} file(s) were skipped as they did '
                'not match the configured include patterns. Use '
                '--include-all-files (or -a) to include all files in the scan.'
            )
    else:
        skipped_message = None

    if metrics.failed_files > 0:
        failed_message = (
                f'Processing of {metrics.failed_files} file(s) failed'
            )
    else:
        failed_message = None

    return ScanFinishedMessages(
            results_message,
            timeouts_message,
            skipped_message,
            failed_message
        )


def default_scan_finished_handler(
            metrics: ScanMetrics,
            timer: timing.Timer,
        ) -> None:
    """Used as the default ScanFinishedCallback"""
    messages = get_scan_finished_messages(metrics, timer)
    if messages.timeouts:
        log.warning(messages.timeouts)
    if messages.skipped:
        log.warning(messages.skipped)
    if messages.failed:
        log.error(messages.failed)
    log.info(messages.results)
    return messages


class ScanWorkerPool:

    def __init__(
                self,
                size: int,
                work_queue: Queue,
                event_queue: Queue,
                matcher: Matcher,
                metrics: ScanMetrics,
                timer: timing.Timer,
                progress_receiver: Optional[ProgressReceiverCallback] = None,
                chunk_size: int = DEFAULT_CHUNK_SIZE,
                scanned_content_limit: Optional[int] = None,
                use_log_events: bool = False,
                allow_io_errors: bool = False,
                debug: bool = False,
                logging_initializer: Callable[[], None] = False,
                profiler: Optional[Profiler] = None,
                direct_io: bool = False
            ):
        self.size = size
        self._matcher = matcher
        self._work_queue = work_queue
        self._event_queue = event_queue
        self.metrics = metrics
        self._timer = timer
        self._progress_receiver = progress_receiver
        self._chunk_size = chunk_size
        self._scanned_content_limit = scanned_content_limit
        self._started = False
        self._use_log_events = use_log_events
        self._allow_io_errors = allow_io_errors
        self._debug = debug
        self._logging_initializer = logging_initializer
        self._profiler = profiler
        self._direct_io = direct_io
        self._completed = False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.stop()
        else:
            self.terminate()

    def has_progress_receiver(self) -> bool:
        return self._progress_receiver is not None

    def _send_progress_update(self) -> None:
        if self.has_progress_receiver():
            update = ScanProgressUpdate(
                    elapsed_time=self._timer.get_elapsed(),
                    metrics=self.metrics
                )
            self._progress_receiver(update)
            self._progress_timer.reset()

    def _is_progress_update_due(self) -> bool:
        if self._progress_timer is None:
            return False
        elapsed = self._progress_timer.get_elapsed(
                timing.unit_milliseconds
            )
        return elapsed >= PROGRESS_UPDATE_INTERVAL

    def _initialize_worker(self, index: int) -> ScanWorker:
        worker = ScanWorker(
                index,
                self._status,
                self._work_queue,
                self._event_queue,
                self._matcher,
                self._chunk_size,
                self._scanned_content_limit,
                self._use_log_events,
                self._allow_io_errors,
                self._logging_initializer,
                self._profiler is not None,
                self._direct_io
            )
        worker.start()
        return worker

    def start(self):
        if self._started:
            raise ScanningException('Worker pool has already been started')
        self._status = Value(c_uint, Status.LOCATING_FILES)
        self._workers = []
        if self.has_progress_receiver():
            self._progress_timer = timing.Timer()
            self._monitor = ScanProgressMonitor(
                    self._status,
                    self._event_queue
                )
            self._monitor.start()
            self._send_progress_update()
        else:
            self._progress_timer = None
            self._monitor = None
        for i in range(self.size):
            worker = self._initialize_worker(i)
            self._workers.append(worker)
        self._started = True

    def _assert_started(self):
        if not self._started:
            raise ScanningException('Worker pool has not been started')

    def stop(self):
        self._assert_started()
        for worker in self._workers:
            worker.join()
        if self._monitor is not None:
            self._monitor.join()

    def terminate(self):
        self._assert_started()
        for worker in self._workers:
            worker.terminate()
        if self._monitor is not None:
            self._monitor.terminate()

    def is_complete(self) -> bool:
        self._assert_started()
        for worker in self._workers:
            if not worker.is_complete():
                return False
        return True

    def check_workers(self) -> None:
        for worker in self._workers:
            if worker.exitcode is not None and worker.exitcode != 0:
                if len(worker.last_file.value) == 0:
                    raise Exception(
                            'Worker exited abnormally (code: '
                            f'{worker.exitcode}) before processing any file'
                        )
                else:
                    log.warning(
                            f'Worker exited abnormally (code: '
                            f'{worker.exitcode})  while processing '
                            + os.fsdecode(worker.last_file.value)
                        )
                    log.info(f'Restarting worker {worker.index}...')
                    self._workers[worker.index] = \
                        self._initialize_worker(worker.index)
                    self.metrics.failed_files += 1

    def await_results(
                self,
                result_processor: ScanResultCallback,
                final: bool = True
            ):
        if self._completed:
            return True
        self._assert_started()
        while True:
            try:
                self.check_workers()
                event = self._event_queue.get(
                        block=final,
                        timeout=RESULT_QUEUE_READ_TIMEOUT
                    )
            except queue.Empty:
                if final:
                    continue
                return False
            if event is None:
                log.debug('All workers have completed and all results have '
                          'been processed.')
                self._status.value = Status.COMPLETE
                self._send_progress_update()
                self._completed = True
                return True
            elif event.type == ScanEventType.COMPLETED:
                if event.worker_index != FILE_LOCATOR_WORKER_INDEX:
                    log.debug(f'Worker {event.worker_index} completed')
                else:
                    log.debug("File locator process exited")
                if self.is_complete():
                    self._event_queue.put(None)
            elif event.type == ScanEventType.FILE_PROCESSED:
                result = ScanResult(
                        event.data['path'],
                        event.data['length'],
                        event.data['matches'],
                        event.data['timeouts']
                    )
                if result.get_timeout_count() > 0:
                    log.warning(
                            'The following signatures timed out while '
                            'processing ' + os.fsdecode(result.path) + ':' +
                            ', '.join({str(i) for i in result.timeouts})
                        )
                self.metrics.record_result(
                        event.worker_index,
                        result
                    )
                result_processor(result)
            elif event.type == ScanEventType.FILE_QUEUE_EMPTIED:
                self._status.value = Status.PROCESSING_FILES
            elif event.type == ScanEventType.EXCEPTION:
                exception = event.data['exception']
                detail = str(exception)
                if self._debug:
                    detail += '\n' + exception.trace
                    detail = detail.strip()
                log.warning(
                        f'Exception occurred during scanning: {detail}'
                    )
            elif event.type == ScanEventType.FATAL_EXCEPTION:
                self._status.value = Status.FAILED
                self.terminate()
                raise event.data['exception']
            elif event.type == ScanEventType.PROGRESS_UPDATE:
                if self._is_progress_update_due():
                    self._send_progress_update()
            elif event.type == ScanEventType.LOG_MESSAGE:
                message: str = event.data['message']
                log.log(event.data['level'], message)
            elif event.type == ScanEventType.PROFILE_EVENT:
                if self._profiler is not None:
                    self._profiler.add_event(event.data)
        return False

    def is_failed(self) -> bool:
        return self._status.value == Status.FAILED


class Scanner:

    def __init__(self, options: Options):
        self.options = options
        self.failed = 0
        self.active = []

    def _handle_worker_error(self, error: Exception):
        self.failed += 1
        raise error

    def _initialize_matcher(self) -> Matcher:
        engine = self.options.match_engine
        options = dataclasses.replace(self.options.match_engine_options)
        options.lazy = not USES_FORK
        return engine.create_matcher(options)

    def scan(
                self,
                result_processor: ScanResultCallback,
                progress_receiver: Optional[ProgressReceiverCallback] = None,
                scan_finished_handler: ScanFinishedCallback =
                default_scan_finished_handler,
                use_log_events: bool = False
            ) -> ScanMetrics:
        """Run a scan"""
        timer = timing.Timer()
        profiler = Profiler() if self.options.profile else None
        event_queue = Queue(MAX_PENDING_RESULTS)
        file_locator_process = FileLocatorProcess(
                file_filter=self.options.file_filter,
                use_log_events=use_log_events,
                event_queue=event_queue if (
                        use_log_events or profiler is not None
                    ) else None,
                allow_io_errors=self.options.allow_io_errors,
                logging_initializer=self.options.logging_initializer,
                profile=profiler is not None
            )
        file_locator_process.start()
        self.active.append(file_locator_process)
        worker_count = self.options.workers
        if worker_count < 1:
            raise ScanConfigurationException(
                    'Scans require at least one worker'
                )
        log.debug("Using " + str(worker_count) + " worker(s)...")
        matcher = self._initialize_matcher()
        metrics = ScanMetrics(worker_count)
        with ScanWorkerPool(
                    size=worker_count,
                    work_queue=file_locator_process.output_queue,
                    event_queue=event_queue,
                    matcher=matcher,
                    metrics=metrics,
                    timer=timer,
                    progress_receiver=progress_receiver,
                    chunk_size=self.options.chunk_size,
                    scanned_content_limit=self.options.scanned_content_limit,
                    use_log_events=use_log_events,
                    allow_io_errors=self.options.allow_io_errors,
                    debug=self.options.debug,
                    logging_initializer=self.options.logging_initializer,
                    profiler=profiler,
                    direct_io=self.options.direct_io
                ) as worker_pool:
            def add_path(path: str):
                while not file_locator_process.add_path(path):
                    worker_pool.await_results(result_processor, final=False)
            self.active.append(worker_pool)
            for path in self.options.paths:
                add_path(path)
            if self.options.path_source is not None:
                log.debug('Reading input paths...')
                while True:
                    path = self.options.path_source.read_entry()
                    if path is None:
                        break
                    add_path(os.fsencode(path))
            while not file_locator_process.finalize_paths():
                worker_pool.await_results(result_processor, final=False)
            log.debug('Awaiting results...')
            worker_pool.await_results(result_processor)
        timer.stop()
        scan_finished_handler = scan_finished_handler if scan_finished_handler\
            else default_scan_finished_handler
        metrics.skipped_files = file_locator_process.get_skipped_count()
        scan_finished_handler(metrics, timer)
        if profiler is not None:
            profiler.complete()
            if self.options.profile_path is None:
                writer_factory = LogProfileWriterFactory()
            else:
                writer_factory = FileProfileWriterFactory(
                        self.options.profile_path
                    )
            with writer_factory as writer:
                profiler.output_results(writer)
        return (metrics, timer)

    def terminate(self) -> None:
        for active in self.active:
            active.terminate()



import requests
from requests.exceptions import RequestException
from enum import Enum
from typing import Callable, Dict, Type, Optional

from .exceptions import ApiException
from ..util.validation import Validator, DictionaryValidator, ListValidator, \
        AllowedValueValidator, OptionalValueValidator, NumberValidator, \
        ValidationException
from ..intel.vulnerabilities import Vulnerability, ScannerVulnerability, \
        ProductionVulnerability, Software, ProductionSoftware, SoftwareType, \
        VersionRange, Cwe, Cvss, CopyrightInformation, Copyright


DEFAULT_BASE_URL = 'https://www.wordfence.com/api/intelligence/v2'
DEFAULT_TIMEOUT = 30


def get_base_vulnerability_feed_validator() -> Validator:
    return DictionaryValidator(
            validator=DictionaryValidator({
                'id': str,
                'title': str,
                'software': ListValidator(DictionaryValidator({
                        'type': AllowedValueValidator({
                                    'core', 'plugin', 'theme'
                                }),
                        'name': str,
                        'slug': str,
                        'affected_versions': DictionaryValidator(
                            validator=DictionaryValidator({
                                'from_version': str,
                                'from_inclusive': bool,
                                'to_version': str,
                                'to_inclusive': bool
                            })),
                        'patched': bool,
                        'patched_versions': ListValidator(str)
                    })),
                'informational': OptionalValueValidator(bool),
                'references': ListValidator(str),
                'published': OptionalValueValidator(str),
                'copyrights': DictionaryValidator(
                        expected={
                            'message': str,
                        },
                        validator=DictionaryValidator({
                            'notice': str,
                            'license': str,
                            'license_url': str
                        }),
                        allow_empty=True
                    )
            }, optional_keys={'informational'})
        )


def get_production_vulnerability_feed_validator() -> Validator:
    validator = get_base_vulnerability_feed_validator()
    validator.validator.add_field('description', str)
    validator.validator.add_field(
            'cwe',
            OptionalValueValidator(DictionaryValidator({
                'id': int,
                'name': str,
                'description': str
            }))
        )
    validator.validator.add_field(
            'cvss',
            OptionalValueValidator(DictionaryValidator({
                'vector': str,
                'score': NumberValidator(),
                'rating': str
            }))
        )
    validator.validator.add_field('cve', OptionalValueValidator(str))
    validator.validator.add_field('cve_link', OptionalValueValidator(str))
    validator.validator.add_field('researchers', ListValidator(str))
    validator.validator.add_field('updated', OptionalValueValidator(str))
    validator.validator.expected['software'].expected.add_field(
            'remediation',
            str
        )
    return validator


class VulnerabilityParser:

    def __init__(
                self,
                type: Type[Vulnerability],
                software_type: Type[Software] = Software
            ):
        self.type = type
        self.software_type = software_type

    def extract_vulnerability_properties(self, record: dict) -> dict:
        properties = {}
        properties['identifier'] = record['id']
        properties['title'] = record['title']
        if 'informational' in record:
            properties['informational'] = record['informational']
        properties['references'] = record['references']
        properties['published'] = record['published']
        return properties

    def parse_version_range(self, record: dict) -> VersionRange:
        return VersionRange(
                from_version=record['from_version'],
                from_inclusive=record['from_inclusive'],
                to_version=record['to_version'],
                to_inclusive=record['to_inclusive']
            )

    def extract_software_properties(self, record: dict) -> dict:
        properties = {}
        properties['type'] = SoftwareType(record['type'])
        properties['name'] = record['name']
        properties['slug'] = record['slug']
        properties['patched'] = record['patched']
        properties['patched_versions'] = record['patched_versions']
        affected_versions = {}
        for key, affected in record['affected_versions'].items():
            range = self.parse_version_range(affected)
            affected_versions[key] = range
        properties['affected_versions'] = affected_versions
        return properties

    def parse_copyright(self, record: dict) -> Copyright:
        return Copyright(
                notice=record['notice'],
                license=record['license'],
                license_url=record['license']
            )

    def parse_copyright_information(self, record: dict) \
            -> Optional[CopyrightInformation]:
        copyrights = record['copyrights']
        if len(copyrights) > 0:
            info = CopyrightInformation()
            if 'message' in copyrights:
                info.message = copyrights['message']
            for key, copyright in copyrights.items():
                if key == 'message':
                    continue
                info.copyrights[key] = self.parse_copyright(copyright)
            return info
        return None

    def parse(self, record: dict) -> Vulnerability:
        vulnerability = self.type(**self.extract_vulnerability_properties(
                record
            ))
        for software in record['software']:
            vulnerability.software.append(
                        self.software_type(**self.extract_software_properties(
                                software
                            ))
                    )
        vulnerability.copyright_information = self.parse_copyright_information(
                record
            )
        return vulnerability


class ScannerVulnerabilityParser(VulnerabilityParser):

    def __init__(self):
        super().__init__(
                type=ScannerVulnerability
            )

    def parse(self, record: dict) -> ScannerVulnerability:
        return super().parse(record)


class ProductionVulnerabilityParser(VulnerabilityParser):

    def __init__(self):
        super().__init__(
                type=ProductionVulnerability,
                software_type=ProductionSoftware
            )

    def extract_vulnerability_properties(self, record: dict) -> dict:
        properties = super().extract_vulnerability_properties(record)
        properties['description'] = record['description']
        properties['cve'] = record['cve']
        properties['cve_link'] = record['cve_link']
        properties['researchers'] = record['researchers']
        properties['updated'] = record['updated']
        return properties

    def extract_software_properties(self, record: dict) -> dict:
        properties = super().extract_software_properties(record)
        properties['remediation'] = record['remediation']
        return properties

    def parse_cwe(self, record: dict) -> Cwe:
        return Cwe(
                identifier=record['id'],
                name=record['name'],
                description=record['description']
            )

    def parse_cvss(self, record: dict) -> Cvss:
        return Cvss(
                vector=record['vector'],
                score=record['score'],
                rating=record['rating']
            )

    def parse(self, record: dict) -> ProductionVulnerability:
        vulnerability = super().parse(record)
        if record['cwe'] is not None:
            vulnerability.cwe = self.parse_cwe(record['cwe'])
        if record['cvss'] is not None:
            vulnerability.cvss = self.parse_cvss(record['cvss'])
        return vulnerability


class VulnerabilityFeedVariant(Enum):
    SCANNER = (
            'scanner',
            get_base_vulnerability_feed_validator,
            ScannerVulnerabilityParser()
        )
    PRODUCTION = (
            'production',
            get_production_vulnerability_feed_validator,
            ProductionVulnerabilityParser()
        )

    def __init__(
                self,
                path: str,
                validator_factory: Callable[[], Validator],
                parser: VulnerabilityParser
            ):
        self.path = path
        self.validator_factory = validator_factory
        self.validator = None
        self.parser = parser

    def get_validator(self) -> Validator:
        if self.validator is None:
            self.validator = self.validator_factory()
        return self.validator

    @classmethod
    def for_path(cls, path):
        for variant in cls:
            if variant.path == path:
                return variant
        raise ValueError(f'Unrecognized vulnerability feed variant: {path}')


class Client:

    def __init__(
                self,
                base_url: Optional[str] = None,
                timeout: int = DEFAULT_TIMEOUT
            ):
        self.base_url = base_url if base_url is not None else DEFAULT_BASE_URL
        self.timeout = timeout

    def _build_url(self, path: str) -> str:
        return self.base_url.rstrip('/') + path

    def fetch_vulnerability_feed(
                self,
                variant: VulnerabilityFeedVariant
            ) -> Dict[str, Vulnerability]:
        url = self._build_url(f'/vulnerabilities/{variant.path}')
        try:
            response = requests.get(url, timeout=self.timeout)
            response.raise_for_status()
            data = response.json()
            variant.get_validator().validate(data)
            vulnerabilities = {}
            for key, record in data.items():
                vulnerabilities[key] = (variant.parser.parse(record))
            return vulnerabilities
        except RequestException as e:
            raise ApiException('Wordfence Intelligence API request failed') \
                from e
        except ValidationException as e:
            raise ApiException(
                        'Wordfence Intelligence API response validation failed'
                    ) \
                from e

    def fetch_scanner_vulnerability_feed(
                self
            ) -> Dict[str, ScannerVulnerability]:
        return self.fetch_vulnerability_feed(VulnerabilityFeedVariant.SCANNER)

    def fetch_production_vulnerability_feed(
                self
            ) -> Dict[str, ProductionVulnerability]:
        return self.fetch_vulnerability_feed(
                VulnerabilityFeedVariant.PRODUCTION
            )



from ..subcommands import Subcommand
from ..configurer import MIN_WORKERS


class ConfigureSubcommand(Subcommand):

    def invoke(self) -> int:
        configurer = self.context.configurer
        configurer.overwrite = self.config.overwrite
        configurer.request_license = self.config.request_license
        if self.config.workers is not None \
                and self.config.workers < MIN_WORKERS:
            if self.config.is_from_cli('workers'):
                raise ValueError(
                        'The number of workers cannot be less than '
                        f'{MIN_WORKERS}'
                    )
            self.config.workers = MIN_WORKERS
        configurer.workers = self.config.workers
        configurer.default = self.config.default
        configurer.prompt_for_config()
        return 0


factory = ConfigureSubcommand



from ..exceptions import CliException, ConfigurationException


class VulnScanningException(CliException):
    pass


class VulnScanningConfigurationException(
            VulnScanningException,
            ConfigurationException
        ):
    pass



import os
from typing import Optional, Dict

from .extension import Extension, ExtensionLoader


THEME_HEADER_FIELDS = {
        'Name': 'Theme Name',
        'ThemeURI': 'Theme URI',
        'Description': 'Description',
        'Author': 'Author',
        'AuthorURI': 'Author URI',
        'Version': 'Version',
        'Template': 'Template',
        'Status': 'Status',
        'Tags': 'Tags',
        'TextDomain': 'Text Domain',
        'DomainPath': 'Domain Path',
        'RequiresWP': 'Requires at least',
        'RequiresPHP': 'Requires PHP'
    }


class Theme(Extension):
    pass


class ThemeLoader(ExtensionLoader):

    def __init__(self, directory: str, allow_io_errors: bool = False):
        super().__init__(
                'theme',
                directory=directory,
                header_fields=THEME_HEADER_FIELDS,
                allow_io_errors=allow_io_errors
            )

    def _initialize_extension(
                self,
                slug: str,
                version: Optional[str],
                header: Dict[str, str],
                path: bytes
            ):
        return Theme(
                slug=slug,
                version=version,
                header=header,
                path=path
            )

    def _process_entry(self, entry: os.DirEntry) -> Optional[Theme]:
        if not entry.is_dir():
            return None
        path = os.path.join(entry.path, b'style.css')
        if not os.path.isfile(path):
            return None
        slug = os.fsdecode(entry.name)
        return self.load(slug, path, base_path=entry.path)



import re
import os.path
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Dict, Optional, Union, Set, Callable, Generator

from ..util.versioning import PhpVersion, compare_php_versions
from ..util.url import Url
from ..wordpress.site import WordpressSite
from ..wordpress.extension import Extension
from ..wordpress.plugin import Plugin
from ..wordpress.theme import Theme


VERSION_ANY = '*'


@dataclass
class VersionRange:
    from_version: str
    from_inclusive: bool
    to_version: str
    to_inclusive: bool

    def includes(self, version: Union[PhpVersion, str]) -> bool:
        from_result = compare_php_versions(self.from_version, version)
        if not (self.from_version == VERSION_ANY or
                from_result == -1 or
                (self.from_inclusive and from_result == 0)):
            return False
        to_result = compare_php_versions(self.to_version, version)
        if not (self.to_version == VERSION_ANY or
                to_result == 1 or
                (self.to_inclusive and to_result == 0)):
            return False
        return True


class SoftwareType(str, Enum):
    CORE = 'core'
    PLUGIN = 'plugin'
    THEME = 'theme'


@dataclass
class ScannableSoftware:
    type: SoftwareType
    slug: str
    version: bytes
    scan_path: Optional[str]

    def get_key(self) -> str:
        return f'{self.type.value}-{self.slug}-{self.version}'


@dataclass
class Software:
    type: SoftwareType
    name: str
    slug: str
    affected_versions: Dict[str, VersionRange] = field(default_factory=dict)
    patched: bool = False
    patched_versions: List[str] = field(default_factory=list)


@dataclass
class Copyright:
    notice: str
    license: str
    license_url: str


@dataclass
class CopyrightInformation:
    message: Optional[str] = None
    copyrights: Dict[str, Copyright] = field(default_factory=dict)


@dataclass
class Vulnerability:
    identifier: str
    title: str
    software: List[Software] = field(default_factory=list)
    informational: bool = False
    references: List[str] = field(default_factory=list)
    published: Optional[str] = None
    copyright_information: Optional[CopyrightInformation] = None

    def get_wordfence_link(self) -> Optional[str]:
        for url in self.references:
            try:
                url_object = Url(url)
                if url_object.get_hostname() == 'www.wordfence.com':
                    url_object.set_query_parameter('source', 'cli-scan')
                    return str(url_object)
            except ValueError:
                continue
        return None

    def get_matched_software(self, scannable: ScannableSoftware) -> Software:
        for software in self.software:
            if software.type != scannable.type \
                    or software.slug != scannable.slug:
                continue
            for affected in software.affected_versions.values():
                if affected.includes(scannable.version):
                    return software


@dataclass
class ScannerVulnerability(Vulnerability):
    pass


@dataclass
class Cwe:
    identifier: int
    name: str
    description: str


@dataclass
class Cvss:
    vector: str
    score: Union[float, int]
    rating: str


@dataclass
class ProductionSoftware(Software):
    remediation: str = ''


@dataclass
class ProductionVulnerability(Vulnerability):
    software: List[ProductionSoftware] = field(default_factory=list)
    description: str = ''
    cwe: Optional[Cwe] = None
    cvss: Optional[Cvss] = None
    cve: Optional[str] = None
    cve_link: Optional[str] = None
    researchers: List[str] = field(default_factory=list)
    updated: Optional[str] = None


SLUG_WORDPRESS = 'wordpress'


class VulnerabilityIndex:

    def __init__(self, vulnerabilities: Dict[str, Vulnerability]):
        self.vulnerabilities = vulnerabilities
        self.id_map = {}
        self.cve_map = {}
        self._initialize_index(vulnerabilities)

    def _add_vulnerability_to_index(
                self,
                vulnerability: Vulnerability
            ) -> None:
        self.id_map[vulnerability.identifier.casefold()] = \
            vulnerability.identifier
        if hasattr(vulnerability, 'cve') and vulnerability.cve is not None:
            self.cve_map[vulnerability.cve.casefold()] = \
                vulnerability.identifier
        for software in vulnerability.software:
            type_index = self.index[software.type]
            if software.slug not in type_index:
                type_index[software.slug] = []
            software_index = type_index[software.slug]
            for version_range in software.affected_versions.values():
                software_index.append(
                        (
                            version_range,
                            vulnerability.identifier
                        )
                    )

    def _initialize_index(self, vulnerabilities: Dict[str, Vulnerability]):
        self.index = {}
        for type in SoftwareType:
            self.index[type] = {}
        for vulnerability in vulnerabilities.values():
            self._add_vulnerability_to_index(vulnerability)

    def get_vulnerabilities(
                self,
                software_type: SoftwareType,
                slug: str,
                version: str
            ) -> Dict[str, Vulnerability]:
        vulnerabilities = {}
        type_index = self.index[software_type]
        if slug in type_index:
            software_index = type_index[slug]
            for version_range, identifier in software_index:
                if version_range.includes(version):
                    vulnerabilities[identifier] = \
                            self.vulnerabilities[identifier]
        return vulnerabilities

    def get_core_vulnerabilties(
                self,
                version: str
            ) -> Dict[str, Vulnerability]:
        return self.get_vulnerabilities(
                SoftwareType.CORE,
                SLUG_WORDPRESS,
                version
            )

    def get_plugin_vulnerabilities(
                self,
                slug: str,
                version: str
            ) -> Dict[str, Vulnerability]:
        return self.get_vulnerabilities(
                SoftwareType.PLUGIN,
                slug,
                version
            )

    def get_theme_vulnerabilities(
                self,
                slug: str,
                version: str
            ) -> Dict[str, Vulnerability]:
        return self.get_vulnerabilities(
                SoftwareType.THEME,
                slug,
                version
            )

    def includes_vulnerability(self, identifier: str) -> bool:
        casefolded = identifier.casefold()
        return casefolded in self.vulnerabilities or casefolded in self.cve_map


CVE_PATTERN = re.compile(r'^CVE-(199\d|20\d{2})-\d{4,}$', re.IGNORECASE)


def is_cve_id(value: str) -> bool:
    return CVE_PATTERN.match(value) is not None


class VulnerabilityFilter:

    def __init__(
                self,
                excluded: Set[str],
                included: Set[str],
                informational: bool = False
            ):
        self.filtered_ids = set(included) | set(excluded)
        self.excluded = self._make_case_insensitive(excluded)
        self.included = self._make_case_insensitive(included)
        self.informational = informational

    def _make_case_insensitive(self, vulnerability_set: Set[str]) -> Set[str]:
        return {identifier.casefold() for identifier in vulnerability_set}

    def _contains_vulnerability(
                self,
                vulnerability_set: Set[str],
                vulnerability: Vulnerability
            ) -> bool:
        if vulnerability.identifier.casefold() in vulnerability_set:
            return True
        if hasattr(vulnerability, 'cve') \
                and vulnerability.cve is not None \
                and vulnerability.cve.casefold() in vulnerability_set:
            return True
        return False

    def allows(self, vulnerability: Vulnerability) -> bool:
        if self._contains_vulnerability(self.excluded, vulnerability):
            return False
        if len(self.included) and \
                not self._contains_vulnerability(self.included, vulnerability):
            return False
        if vulnerability.informational and not self.informational:
            return False
        return True

    def filter(
                self,
                vulnerabilities: Dict[str, Vulnerability]
            ) -> Dict[str, Vulnerability]:
        return {
                identifier: vulnerability for identifier, vulnerability
                in vulnerabilities.items() if self.allows(vulnerability)
            }

    def get_invalid_ids(
                self,
                index: VulnerabilityIndex
            ) -> Generator[None, None, str]:
        for identifier in self.filtered_ids:
            if not index.includes_vulnerability(identifier):
                yield identifier


DEFAULT_FILTER = VulnerabilityFilter(
        excluded={},
        included={},
        informational=False
    )


class AlreadyScannedException(Exception):
    pass


class VulnerabilityScanner:

    def __init__(
                self,
                index: VulnerabilityIndex,
                filter: VulnerabilityFilter = DEFAULT_FILTER
            ):
        self.index = index
        self.filter = filter
        self.vulnerabilities = {}
        self.affected = {}
        self.callbacks = []
        self.scan_paths = set()

    def register_result_callback(
                    self,
                    callback: Callable[
                        [ScannableSoftware, Dict[str, Vulnerability]],
                        None
                    ]
                ) -> None:
        self.callbacks.append(callback)

    def _trigger_callbacks(
                self,
                software: ScannableSoftware,
                vulnerabilities: Dict[str, Vulnerability]
            ) -> None:
        for callback in self.callbacks:
            callback(software, vulnerabilities)

    def add_scan_path(self, path: str) -> None:
        realpath = os.path.realpath(path)
        if realpath in self.scan_paths:
            raise AlreadyScannedException(f'{path} has already been scanned')
        self.scan_paths.add(realpath)

    def scan(self, software: ScannableSoftware) -> Dict[str, Vulnerability]:
        vulnerabilities = self.index.get_vulnerabilities(
                software.type,
                software.slug,
                software.version
            )
        vulnerabilities = self.filter.filter(vulnerabilities)
        self._trigger_callbacks(software, vulnerabilities)
        self.vulnerabilities.update(vulnerabilities)
        for identifier in vulnerabilities:
            if identifier not in self.affected:
                self.affected[identifier] = []
            self.affected[identifier].append(software)
        return vulnerabilities

    def scan_core(
                self,
                version: bytes,
                scan_path: Optional[str]
            ) -> Dict[str, Vulnerability]:
        return self.scan(
                ScannableSoftware(
                    type=SoftwareType.CORE,
                    slug=SLUG_WORDPRESS,
                    version=version,
                    scan_path=scan_path
                )
            )

    def scan_site(
                self,
                site: WordpressSite,
                scan_path: Optional[str] = None
            ) -> Dict[str, Vulnerability]:
        return self.scan_core(site.get_version(), scan_path)

    def scan_extension(
                self,
                extension: Extension,
                type: SoftwareType,
                scan_path: Optional[str] = None
            ) -> Dict[str, Vulnerability]:
        return self.scan(
                ScannableSoftware(
                    type=type,
                    slug=extension.slug,
                    version=extension.version,
                    scan_path=scan_path
                )
            )

    def scan_plugin(
                self,
                plugin: Plugin,
                scan_path: Optional[str] = None
            ) -> Dict[str, Vulnerability]:
        return self.scan_extension(plugin, SoftwareType.PLUGIN, scan_path)

    def scan_theme(
                self,
                theme: Theme,
                scan_path: Optional[str] = None
            ) -> Dict[str, Vulnerability]:
        return self.scan_extension(theme, SoftwareType.THEME, scan_path)

    def get_vulnerability_count(self) -> int:
        return len(self.vulnerabilities)

    def get_affected_count(self) -> int:
        affected = set()
        for group in self.affected.values():
            for software in group:
                affected.add(software.get_key())
        return len(affected)

    def get_total_count(self) -> int:
        count = 0
        for group in self.affected.values():
            count += len(group)
        return count



import os

from typing import List, IO, Optional, Dict
from email.message import EmailMessage
from email.headerregistry import Address

from wordfence.scanning.scanner import ScanResult
from wordfence.intel.signatures import SignatureSet, Signature
from wordfence.util.html import Tag
from wordfence.util.units import scale_byte_unit
from wordfence.util.encoding import bytes_to_str

from ..reporting import Report, ReportFormat, get_config_options, \
        ReportFormatEnum, ReportColumnEnum, ReportRecord, ReportWriter, \
        ReportManager, ReportColumn, ReportEmail, generate_report_email_html, \
        generate_html_table, \
        REPORT_FORMAT_CSV, REPORT_FORMAT_TSV, REPORT_FORMAT_NULL_DELIMITED, \
        REPORT_FORMAT_LINE_DELIMITED
from ..context import CliContext
from ..email import Mailer
from .progress import ProgressDisplay


class ScanReportColumn(ReportColumnEnum):
    FILENAME = 'filename', lambda record: os.fsdecode(record.result.path)
    SIGNATURE_ID = 'signature_id', lambda record: record.signature.identifier
    SIGNATURE_NAME = 'signature_name', lambda record: record.signature.name
    SIGNATURE_DESCRIPTION = 'signature_description', \
        lambda record: record.signature.description
    MATCHED_TEXT = 'matched_text', \
        lambda record: (
                '' if record.match is None
                else bytes_to_str(record.match)
            )


class HumanReadableWriter(ReportWriter):

    def __init__(self, target: IO, columns: List[ScanReportColumn]):
        super().__init__(target)
        self._columns = columns

    def _get_value(data: List[str], column: str) -> str:
        return

    def _map_data_to_dict(self, data: List[str]) -> dict:
        return {
            column.header: data[index] for index, column
            in enumerate(self._columns)
        }

    def write_row(self, data: List[str]) -> None:
        values = self._map_data_to_dict(data)
        file = None
        signature_id = None
        if 'filename' in values:
            file = values['filename']
        if 'signature_id' in values:
            signature_id = values['signature_id']
        # TODO: Add more custom messages if desired
        if file is not None:
            if signature_id is not None:
                self._target.write(
                        f"File at {file} matched signature {signature_id}"
                    )
            else:
                self._target.write(
                        f"File {file} matched a signature"
                    )
        else:
            self._target.write(
                    "Match found: " + str(values)
                )
        self._target.write("\n")

    def allows_headers(self) -> bool:
        return False


REPORT_FORMAT_HUMAN = ReportFormat(
        'human',
        lambda stream, columns: HumanReadableWriter(stream, columns),
        allows_headers=False
    )


class ScanReportFormat(ReportFormatEnum):
    CSV = REPORT_FORMAT_CSV
    TSV = REPORT_FORMAT_TSV
    NULL_DELIMITED = REPORT_FORMAT_NULL_DELIMITED
    LINE_DELIMITED = REPORT_FORMAT_LINE_DELIMITED
    HUMAN = REPORT_FORMAT_HUMAN


class ScanReportRecord(ReportRecord):

    def __init__(
                self,
                result: ScanResult,
                signature: Signature,
                match: bytes
            ):
        self.result = result
        self.signature = signature
        self.match = match


class ScanReport(Report):

    def __init__(
                self,
                format: ScanReportFormat,
                columns: List[ScanReportColumn],
                signature_set: SignatureSet,
                email_addresses: List[str],
                mailer: Optional[Mailer],
                write_headers: bool = False
            ):
        super().__init__(
                format=format,
                columns=columns,
                email_addresses=email_addresses,
                mailer=mailer,
                write_headers=write_headers
            )
        self.signature_set = signature_set
        self.metrics = None

    def add_result(self, result: ScanResult) -> None:
        records = []
        for signature_id, match in result.matches.items():
            signature = self.signature_set.get_signature(signature_id)
            record = ScanReportRecord(
                    result=result,
                    signature=signature,
                    match=match
                )
            records.append(record)
        self.write_records(records)

    def generate_email(
                self,
                recipient: Address,
                attachments: Dict[str, str],
                hostname: str
            ) -> EmailMessage:

        suspicious_count = self.metrics.get_total_matches()
        total_count = self.metrics.get_total_count()
        byte_unit = scale_byte_unit(self.metrics.get_total_bytes())
        elapsed = round(self.timer.get_elapsed())

        results = {
                'Suspicious Files': suspicious_count,
                'Scanned Files': total_count,
                'Bytes Processed': byte_unit,
                'Elapsed Time': f'{elapsed} second(s)'
            }

        content = Tag('div')

        plain = ('Suspicious files were found by Wordfence CLI during a '
                 'malware scan.\n\n')

        content.append(Tag('p').append(plain))

        for label, value in results.items():
            plain += f'{label}: {value}\n'

        table = generate_html_table(results)
        content.append(table)

        document = generate_report_email_html(
                content,
                'Malware Scan Results',
                hostname
            )

        return ReportEmail(
                recipient=recipient,
                subject=f'Malware Scan Results for {hostname}',
                plain_content=plain,
                html_content=document.to_html()
            )


SCAN_REPORT_CONFIG_OPTIONS = get_config_options(
        ScanReportFormat,
        ScanReportColumn
    )


class ScanReportManager(ReportManager):

    def __init__(
                self,
                context: CliContext,
                signature_set: SignatureSet
            ):
        super().__init__(
                formats=ScanReportFormat,
                columns=ScanReportColumn,
                context=context,
                read_stdin=context.config.read_stdin,
                input_delimiter=context.config.file_list_separator,
                binary_input=True
            )
        self.signature_set = signature_set
        self.progress = None

    def set_progress_display(self, progress: ProgressDisplay) -> None:
        self.progress = progress

    def _instantiate_report(
                self,
                format: ReportFormat,
                columns: List[ReportColumn],
                email_addresses: List[str],
                mailer: Optional[Mailer],
                write_headers: bool
            ) -> ScanReport:
        return ScanReport(
                format,
                columns,
                self.signature_set,
                email_addresses,
                mailer,
                write_headers
            )

    def _get_stdout_target(self) -> IO:
        if self.progress is not None:
            return self.progress.get_output_stream()
        return super()._get_stdout_target()



import os

from ...wordpress.remediator import Remediator, Noc1RemediationSource
from ...logging import log
from ..subcommands import Subcommand
from ..exceptions import ConfigurationException
from .reporting import RemediationReportManager, RemediationReport


class RemediateSubcommand(Subcommand):

    def prepare(self) -> None:
        self.remediator = Remediator(
                Noc1RemediationSource(self.context.get_noc1_client())
            )

    def process_path(self, path: bytes, report: RemediationReport) -> None:
        log.debug(f'Attempting to remediate {path}...')
        for result in self.remediator.remediate(os.fsencode(path)):
            report.add_result(result)

    def invoke(self) -> int:
        report_manager = RemediationReportManager(self.context)
        io_manager = report_manager.get_io_manager()
        with report_manager.open_output_file() as output_file:
            report = report_manager.initialize_report(output_file)
            for path in self.config.trailing_arguments:
                self.process_path(path, report)
            if io_manager.should_read_stdin():
                reader = io_manager.get_input_reader()
                for path in reader.read_all_entries():
                    self.process_path(path, report)
            if self.remediator.input_count == 0 and \
                    self.context.requires_input(self.config.require_path):
                raise ConfigurationException(
                        'At least one path to remediate must be specified'
                    )
            report.complete()
            if report.counts.remediated == report.counts.total:
                log.info(
                        f'{report.counts.remediated} file(s) were successfully'
                        ' remediated'
                    )
            else:
                log.error(
                    f'{report.counts.remediated} of {report.counts.total} '
                    'file(s) were successfully remediated, '
                    f'{report.counts.unsuccessful} file(s) could not be '
                    'remediated'
                )
                return 1
        return 0


factory = RemediateSubcommand



import abc
import base64
import json
from dataclasses import dataclass, fields
from enum import Enum
from functools import lru_cache
from typing import Optional, Any, Dict, Set, Tuple, Type, Callable, Union
from .typing import ConfigDefinitions

# see wordfence/config/__init__.py for special handling of reserved names
invalid_config_item_names: Set[str] = {
    'subcommand',
    'trailing_arguments'
}


class Context(Enum):
    ALL = 1
    """a config item that is available in both the CLI and INI contexts"""
    CLI = 2
    """a config item that is only available in the CLI context (not INI)"""
    CONFIG = 3
    """a config item that is only available in the INI context (not CLI)"""


class ArgumentType(Enum):
    ARGUMENT = 1
    """No names, just ordered CLI values"""
    FLAG = 2
    """boolean values set by name (no value) -- inverts the default value when
    provided"""
    OPTIONAL_FLAG = 3
    """boolean values set by name with an optional third state (None)
    representing the absence of a value"""
    OPTION = 4
    """required the option name plus a value"""
    OPTION_REPEATABLE = 5
    """an option that can be repeated multiple times with different values"""


@dataclass(frozen=True)
class ReferenceToken:
    """Instantiate a new instance to use the `x is y` language construct to
    determine if other instances point to the same token"""
    pass


not_set_token = ReferenceToken()


@dataclass(frozen=True)
class ConfigItemMeta:
    valid_options: Optional[Tuple[str]] = None
    multiple: Optional[bool] = None
    separator: Optional[str] = None
    value_type: Union[Type, Callable] = str
    accepts_file: bool = False
    accepts_directory: bool = False

    def accepts_paths(self) -> bool:
        return self.accepts_file or self.accepts_directory


@dataclass(frozen=True)
class ConfigItemDefinition:
    name: str
    property_name: str
    description: str
    context: Context
    argument_type: ArgumentType
    default: Any
    hidden: bool = False
    short_name: Optional[str] = None
    meta: Optional[ConfigItemMeta] = None
    category: str = 'General Options'

    @staticmethod
    def clean_argument_dict(source: Dict[str, Any]) -> Dict[str, Any]:
        return {key: value for key, value in source.items() if
                key in get_data_item_fields()}

    def has_options_list(self) -> bool:
        return True if self.meta and self.meta.valid_options else False

    def has_separator(self) -> bool:
        return True if self.meta and self.meta.separator else False

    def is_flag(self) -> bool:
        return self.argument_type == ArgumentType.FLAG \
            or self.argument_type == ArgumentType.OPTIONAL_FLAG

    def accepts_value(self) -> bool:
        return not self.is_flag()

    def get_value_type(self):
        if self.is_flag():
            return bool
        if not self.meta:
            return str
        return_type = self.meta.value_type
        if not return_type:
            if self.meta.accepts_paths():
                return bytes
            raise ValueError(
                f"Specified type not in the allow list: {self.meta.value_type}"
                )
        return return_type

    def accepts_paths(self) -> bool:
        return self.meta and self.meta.accepts_paths()

    @classmethod
    def from_dict(cls, source: dict):
        # The property name is always derived from the configuration's "name"
        # value. Any "property_name" value specified in the configuration is
        # ignored.
        source['property_name'] = source['name'].replace('-', '_')

        is_optional_flag = \
            ArgumentType.OPTIONAL_FLAG == source['argument_type']
        is_flag = (
                is_optional_flag or
                ArgumentType.FLAG == source['argument_type']
            )

        if source.get('default_type', None) == 'base64':
            if 'default' not in source:
                raise ValueError(
                    "When base64 default type is specified, a default value "
                    "must be present")
            source['default'] = base64.b64decode(source['default'])

        if 'default' not in source:
            source['default'] = not_set_token
        # convert enums
        source['context'] = source['context'] if isinstance(source['context'],
                                                            Context) else \
            Context[source['context']]
        source['argument_type'] = source['argument_type'] if isinstance(
            source['argument_type'], ArgumentType) else \
            ArgumentType[
                source['argument_type']]

        # convert the meta dict to an object to make it hashable
        if source.get('meta', False):
            # convert lists to tuples to make them hashable
            if source['meta'].get('valid_options', False):
                source['meta']['valid_options'] = tuple(
                    source['meta']['valid_options'])
            # set flags to booleans types
            # if another type is not already defined
            if not_set_token is source['meta'].get('value_type',
                                                   not_set_token) and is_flag:
                source['meta']['value_type'] = 'bool'
            source['meta'] = ConfigItemMeta(**source['meta'])
        else:
            source['meta'] = ConfigItemMeta()

        # sanity check
        if is_flag and not (
                    isinstance(source['default'], bool) or
                    (is_optional_flag and source['default'] is None)
                ):
            raise ValueError(
                f"Flag {source['name']} has a non-boolean value type defined: "
                f"{type(source['default'])}")
        return cls(**ConfigItemDefinition.clean_argument_dict(source))

    @classmethod
    def from_json(cls, source: str):
        return cls.from_dict(json.loads(source))


@dataclass(frozen=True)
class ConfigValue:
    definition: ConfigItemDefinition
    value: Any


class CanonicalValueExtractorInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (callable(subclass.get_canonical_value),
                callable(subclass.is_valid_source),
                callable(subclass.assert_is_valid_source))

    @abc.abstractmethod
    def is_valid_source(self, source: Any) -> bool:
        """Validate the source is supported"""
        raise NotImplementedError

    def assert_is_valid_source(self, source: Any) -> None:
        if not self.is_valid_source(source):
            raise ValueError(f"Invalid configuration source: {type(source)}")

    @abc.abstractmethod
    def get_canonical_value(self, definition: ConfigItemDefinition,
                            source: Any) -> Any:
        """Return the canonical configuration value as stored in the
        configuration source"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_context(self) -> Context:
        raise NotImplementedError


@lru_cache(maxsize=1)
def get_data_item_fields() -> Set[str]:
    return set([x.name for x in fields(ConfigItemDefinition)])


def config_definitions_to_config_map(
            config_definitions: ConfigDefinitions
        ) -> Dict[str, ConfigItemDefinition]:
    result: Dict[str, ConfigItemDefinition] = {}
    used_short_names: Set[str] = set()
    implied_names: Set[str] = set()
    for name, value in config_definitions.items():
        value['name'] = name
        config_item = ConfigItemDefinition.from_dict(value)
        if config_item.name in result:
            raise KeyError(
                f"The name {json.dumps(config_item.name)} has already been "
                f"loaded")
        if config_item.name in implied_names:
            raise KeyError(
                f"A configured flag has already claimed "
                f"{json.dumps(config_item.name)} as an implied name")
        if config_item.short_name:
            if config_item.short_name in used_short_names:
                raise KeyError(
                    f"The short name {json.dumps(config_item.short_name)} "
                    f"has already been loaded")
            else:
                used_short_names.add(config_item.short_name)
        if config_item.is_flag():
            implied_name = f'no-{config_item.name}'
            if implied_name in implied_names:
                raise KeyError(
                    f"Another option has already taken the implied name "
                    f"{json.dumps(implied_name)}")
            implied_names.add(implied_name)

        if config_item.name in invalid_config_item_names:
            raise KeyError(
                f"The name {json.dumps(config_item.name)} is reserved")
        result[config_item.name] = config_item
    return result


def merge_config_maps(
            a: Dict[str, ConfigItemDefinition],
            b: Dict[str, ConfigItemDefinition]
        ) -> Dict[str, ConfigItemDefinition]:
    return {**a, **b}



from wordfence.util.validation import ListValidator, DictionaryValidator, \
    OptionalValueValidator
from typing import Optional, Set, List
import json


class DatabaseRule:

    def __init__(
                self,
                identifier: int,
                tables: Optional[Set[str]] = None,
                condition: Optional[str] = None,
                description: Optional[str] = None
            ):
        self.identifier = identifier
        self.tables = tables
        self.condition = condition
        self.description = description

    def __hash__(self):
        return hash(self.identifier)

    def __eq__(self, other) -> bool:
        return (
                type(other) is type(self)
                and other.identifier == self.identifier
            )


class DatabaseRuleSet:

    def __init__(self):
        self.rules = {}
        self.table_rules = {}
        self.global_rules = set()

    def add_rule(self, rule: DatabaseRule) -> None:
        if rule.identifier in self.rules:
            raise Exception('Duplicate rule ID: {rule.identifier}')
        self.rules[rule.identifier] = rule
        if rule.tables is None:
            self.global_rules.add(rule)
        else:
            for table in rule.tables:
                if table not in self.table_rules:
                    self.table_rules[table] = set()
                self.table_rules[table].add(rule)

    def remove_rule(self, rule_id: int) -> None:
        try:
            rule = self.rules.pop(rule_id)
            if rule.tables is None:
                self.global_rules.discard(rule)
            else:
                for table in rule.tables:
                    if table in list(self.table_rules.keys()):
                        table_rules = self.table_rules[table]
                        table_rules.discard(rule)
                        if len(table_rules) == 0:
                            del self.table_rules[table]
        except KeyError:
            pass  # Rule doesn't exist, no need to remove

    def get_rules(self, table: str) -> List[DatabaseRule]:
        rules = []
        try:
            rules.extend(self.table_rules[table])
        except KeyError:
            pass  # There are no table rules
        rules.extend(self.global_rules)
        return rules

    def get_targeted_tables(self) -> List[str]:
        return self.table_rules.keys()

    def get_rule(self, identifier: int) -> DatabaseRule:
        return self.rules[identifier]

    def filter_rules(
                self,
                included: Optional[Set[int]] = None,
                excluded: Optional[Set[int]] = None
            ):
        if included is not None:
            for rule_id in list(self.rules.keys()):
                if rule_id not in included:
                    self.remove_rule(rule_id)
        if excluded is not None:
            for rule_id in excluded:
                self.remove_rule(rule_id)


JSON_VALIDATOR = ListValidator(
        DictionaryValidator({
                'id': int,
                'tables': ListValidator(str),
                'condition': str,
                'description': OptionalValueValidator(str)
            }, optional_keys={'description'})
    )


def parse_database_rules(
            data,
            pre_validated: bool = False,
            rule_set: Optional[DatabaseRuleSet] = None
        ) -> DatabaseRuleSet:
    if not pre_validated:
        JSON_VALIDATOR.validate(data)
    if rule_set is None:
        rule_set = DatabaseRuleSet()
    for rule_data in data:
        rule = DatabaseRule(
                identifier=rule_data['id'],
                tables=rule_data['tables'],
                condition=rule_data['condition'],
                description=rule_data['description']
            )
        rule_set.add_rule(rule)
    return rule_set


def load_database_rules(
            path: bytes,
            rule_set: Optional[DatabaseRuleSet] = None
        ) -> DatabaseRuleSet:
    with open(path, 'rb') as file:
        data = json.load(file)
    return parse_database_rules(data, rule_set=rule_set)



class WordpressException(Exception):
    pass


class ExtensionException(WordpressException):
    pass


class WordpressDatabaseException(Exception):

    def __init__(self, database, message):
        self.database = database



from typing import Optional, Union

from ..api.licensing import License, LicenseSpecific, to_license
from ..api.exceptions import ApiException
from ..api import noc1
from ..util.caching import NoCachedValueException, InvalidCachedValueException
from .context import CliContext


CACHE_KEY = 'license'
CACHEABLE_TYPES = {
        'wordfence.api.licensing.LicenseSpecific'
    }


class LicenseValidationFailure(Exception):

    def __init__(self, message: str):
        self.message = message


class LicenseManager:

    def __init__(self, context: CliContext):
        self.context = context

    def _create_noc1_client(
                self,
                license: Optional[License] = None
            ) -> noc1.Client:
        return self.context.create_noc1_client(license)

    def request_free_license(self, terms_accepted: bool = False) -> License:
        client = self.context.create_noc1_client()
        return License(client.get_cli_api_key(accept_terms=terms_accepted))

    def validate_license(self, license: Union[License, str]) -> License:
        license = to_license(license)
        client = self.context.create_noc1_client(license)
        try:
            if not client.ping_api_key():
                raise LicenseValidationFailure('Invalid license')
        except ApiException as exception:
            if exception.public_message is None:
                raise LicenseValidationFailure(
                        'License validation failed.'
                    )
            else:
                raise LicenseValidationFailure(
                        f'Invalid license: {exception.public_message}'
                    )
        return license

    def set_license(self, license: Union[License, str]) -> str:
        license = to_license(license)
        self.context.cache.put(CACHE_KEY, LicenseSpecific(license))

    def check_license(self) -> License:
        current = self.context.get_license()
        try:
            cached = self.context.cache.get(CACHE_KEY)
            if cached.is_compatible_with_license(current):
                return cached.license
        except NoCachedValueException:
            pass
        except InvalidCachedValueException:
            pass
        self.validate_license(current)
        self.set_license(current)
        return current

    def update_license(self, license: License) -> None:
        self.set_license(license)



import pickle
from io import BytesIO
from typing import Any, Set, Optional


class SerializationException(Exception):
    pass


class ProhibitedTypeException(SerializationException):
    pass


class UnexpectedTypeException(SerializationException):
    pass


class LimitedDeserializer(pickle.Unpickler):

    def __init__(self, data: bytes, allowed: Set[str]):
        super().__init__(BytesIO(data))
        self.allowed = allowed

    def find_class(self, module, name):
        full_name = f'{module}.{name}'
        if full_name in self.allowed:
            return super().find_class(module, name)
        else:
            raise ProhibitedTypeException(f'Global {full_name} is not allowed')


def limited_deserialize(
            data: bytes,
            allowed: Optional[Set[str]] = None,
            expected: Any = None
        ) -> Any:
    if allowed is None:
        allowed = set()
    result = LimitedDeserializer(data, allowed).load()
    if expected is not None and not isinstance(result, expected):
        raise UnexpectedTypeException('Unexpected type: ' + type(expected))
    return result



import re
from dataclasses import dataclass
from enum import Enum

KIBIBYTE = 1024
MEBIBYTE = 1024 * 1024

sizings_map = {
    'b': 1,
    'k': KIBIBYTE,
    'kb': KIBIBYTE,
    'kib': KIBIBYTE,
    'm': MEBIBYTE,
    'mb': MEBIBYTE,
    'mib': MEBIBYTE
}
"""maps suffixes to byte multipliers; k/kb/kib are synonyms, as are m/mb/mib"""


def byte_length(conversion_value: str) -> int:
    match = re.search(r"(\d+)([^0-9].*)", conversion_value)
    if not match:
        raise ValueError("Invalid format for byte length type")
    suffix = match.group(2).lower()
    if not sizings_map.get(suffix, False):
        raise ValueError("Unrecognized byte length suffix")
    return int(match.group(1)) * sizings_map.get(suffix)


class ByteUnit(Enum):
    BYTE = (1, 'B')
    KIBIBYTE = (pow(2, 10), 'KiB')
    MEBIBYTE = (pow(2, 20), 'MiB')
    GIBIBYTE = (pow(2, 30), 'GiB')
    TEBIBYTE = (pow(2, 40), 'TiB')

    def __init__(
                self,
                size: int,
                abbreviation: str
            ):
        self.size = size
        self.abbreviation = abbreviation


@dataclass
class ByteUnitValue:
    value: float
    unit: ByteUnit

    def __str__(self) -> str:
        if self.unit.size == 1:
            value = int(self.value)
        else:
            value = round(self.value, 1)
        return f'{value} {self.unit.abbreviation}'


def scale_byte_unit(byte_count: int) -> ByteUnitValue:
    scaled_unit = ByteUnit.BYTE
    for unit in ByteUnit:
        if byte_count >= unit.size \
                and (
                    scaled_unit is None
                    or unit.size > scaled_unit.size
                ):
            scaled_unit = unit
    scaled_value = byte_count / scaled_unit.size
    return ByteUnitValue(scaled_value, scaled_unit)



import sys
import signal
import logging
import pickle
import tempfile
import os
from multiprocessing import parent_process
from typing import Optional, Tuple

from wordfence import scanning
from wordfence.scanning import filtering
from wordfence.scanning.matching import MatchEngine, MatchEngineOptions, \
    MatchEngineCompilerOptions
from wordfence.util import caching, pcre, vectorscan
from wordfence.util.platform import Platform, UnknownPlatform
from wordfence.intel.signatures import SignatureSet, PrecompiledSignatureSet, \
    deserialize_precompiled_signature_set
from wordfence.util.io import chmod_with_umask
from wordfence.logging import (log, remove_initial_handler,
                               restore_initial_handler)
from ..subcommands import Subcommand
from .reporting import ScanReportManager
from .progress import ProgressDisplay, ProgressException, reset_terminal


screen_handler: Optional[logging.Handler] = None


def revert_progress_changes() -> None:
    global screen_handler
    if screen_handler:
        log.removeHandler(screen_handler)
    restore_initial_handler()
    reset_terminal()


class MalwareScanSubcommand(Subcommand):

    def _filter_signatures(
                self,
                signatures: SignatureSet,
            ) -> None:
        if self.config.include_signatures:
            for identifier in list(signatures.signatures.keys()):
                if identifier not in self.config.include_signatures:
                    signatures.remove_signature(identifier)
            for identifier in self.config.include_signatures:
                if identifier in signatures.signatures:
                    log.debug(f'Including signature: {identifier}')
                else:
                    log.warning(
                                f'Signature {identifier} was not found and '
                                'could not be included'
                            )
        if self.config.exclude_signatures is not None:
            for identifier in self.config.exclude_signatures:
                if signatures.remove_signature(identifier):
                    log.debug(f'Excluded signature {identifier}')
                else:
                    log.warning(
                            f'Signature {identifier} is not in the existing '
                            'set. It will not be used in the scan.'
                        )
        signature_count = len(signatures.signatures)
        log.debug(f'Filtered signature count: {signature_count}')

    def _get_pre_compiled_signatures(
                self,
                match_engine: MatchEngine
            ) -> Optional[PrecompiledSignatureSet]:

        def fetch_pre_compiled() -> Optional[PrecompiledSignatureSet]:
            client = self.context.get_noc1_client()
            try:
                pre_compiled = client.get_precompiled_malware_signatures(
                        Platform.detect(),
                        vectorscan.API_VERSION
                    )
                if pre_compiled is None:
                    log.warning(
                            'No compatible pre-compiled signature set was '
                            'found. Signatures will be compiled locally.'
                        )
                return pre_compiled
            except UnknownPlatform:
                log.warning(
                        'Unable to determine current platform, pre-compiled '
                        'signatures cannot be used'
                    )
                return None

        cacheable = caching.Cacheable(
                f'pre-compiled-signatures-{match_engine.module}',
                fetch_pre_compiled,
                caching.DURATION_ONE_DAY
            )
        return cacheable.get(self.cache)

    def _get_signatures(
                self,
                match_engine: MatchEngine,
                check_precompiled: bool = True
            ) -> SignatureSet:
        supports_pre_compilation = match_engine.supports_pre_compilation()
        if supports_pre_compilation and check_precompiled:
            precompiled = self._get_pre_compiled_signatures(match_engine)
            if precompiled is not None:
                filtered = precompiled.signature_set
                if match_engine.validate_database_source(precompiled.data):
                    self._filter_signatures(filtered)
                    return filtered, precompiled.data
                else:
                    log.warning(
                            'Pre-compiled database source is not valid...'
                            'compiling locally...'
                        )

        def fetch_signatures() -> SignatureSet:
            noc1_client = self.context.get_noc1_client()
            return noc1_client.get_malware_signatures()

        self.cacheable_signatures = caching.Cacheable(
                'signatures',
                fetch_signatures,
                caching.DURATION_ONE_DAY
            )
        signatures = self.cacheable_signatures.get(self.cache)
        self._filter_signatures(signatures)
        return signatures, None

    def _get_file_list_separator(self) -> str:
        if isinstance(self.config.file_list_separator, bytes):
            return self.config.file_list_separator.decode('utf-8')
        return self.config.file_list_separator

    def _initialize_file_filter(self) -> filtering.FileFilter:
        filter = filtering.FileFilter()
        has_include_overrides = False
        if self.config.include_all_files:
            filter.add(filtering.filter_any)
        if self.config.include_files is not None:
            has_include_overrides = True
            for name in self.config.include_files:
                filter.add(filtering.FilenameFilter(os.fsencode(name)))
        if self.config.include_files_pattern is not None:
            has_include_overrides = True
            for pattern in self.config.include_files_pattern:
                try:
                    filter.add(filtering.filter_pattern(os.fsencode(pattern)))
                except filtering.InvalidPatternException:
                    raise Exception(
                            'File inclusion pattern must be a Python regex, '
                            f'received: "{pattern}"'
                        )
        if self.config.exclude_files is not None:
            for name in self.config.exclude_files:
                filter.add(filtering.FilenameFilter(os.fsencode(name)), False)
        if self.config.exclude_files_pattern is not None:
            for pattern in self.config.exclude_files_pattern:
                try:
                    filter.add(
                            filtering.filter_pattern(
                                os.fsencode(pattern)
                            ),
                            False
                        )
                except filtering.InvalidPatternException:
                    raise Exception(
                            'File exclusion pattern must be a Python regex, '
                            f'received: "{pattern}"'
                        )
        if not has_include_overrides:
            filter.add(filtering.filter_php)
            filter.add(filtering.filter_html)
            filter.add(filtering.filter_js)
            if self.config.images:
                filter.add(filtering.filter_images)
        return filter

    def _get_pcre_options(self) -> pcre.PcreOptions:
        return pcre.PcreOptions(
                    caseless=True,
                    match_limit=self.config.pcre_backtrack_limit,
                    match_limit_recursion=self.config.pcre_recursion_limit
                )

    def _get_compiled_signatures_cache_key(
                self,
                match_engine: MatchEngine
            ) -> str:
        return f'compiled-signatures-{match_engine.module}'

    def _get_database_source(
                self,
                match_engine: MatchEngine,
                match_engine_options: MatchEngineOptions,
                force: bool = False,
                generic: bool = True
            ) -> Tuple[Optional[bytes], Optional[caching.Cacheable]]:
        current_hash = match_engine_options.signature_set.get_hash()

        def compile_database():
            compiler_options = MatchEngineCompilerOptions(
                    generic=generic
                )
            compiler = match_engine.get_compiler(compiler_options)
            if compiler is None:
                return None
            compiled = compiler.compile_serializable(
                    match_engine_options.signature_set
                )
            signature_set = PrecompiledSignatureSet(
                    match_engine_options.signature_set,
                    compiled,
                    current_hash,
                    match_engine_options.signature_set.license
                )
            if generic:
                signature_set.clear_license()
            return signature_set

        def is_precompiled_compatible(precompiled):
            return (
                   precompiled is not None and
                   precompiled.signature_hash == current_hash and
                   precompiled.is_supported_version()
               )

        if self.config.pattern_database_path is None:

            def filter_precompiled(precompiled):
                if not is_precompiled_compatible(precompiled):
                    raise caching.InvalidCachedValueException(
                            'Incompatible signature set'
                        )
                return precompiled
            cacheable = caching.Cacheable(
                    self._get_compiled_signatures_cache_key(match_engine),
                    compile_database,
                    filters=[
                        filter_precompiled
                    ]
                )
            if force:
                cacheable.delete(self.cache)
            precompiled = cacheable.get(self.cache)

        else:

            precompiled = None
            database_path = self.config.pattern_database_path
            if not force:
                try:
                    with open(database_path, 'rb') as file:
                        existing = file.read()
                        if len(existing) > 0:
                            existing = \
                                deserialize_precompiled_signature_set(
                                    existing
                                )
                            if is_precompiled_compatible(existing):
                                precompiled = existing
                except FileNotFoundError:
                    pass

            if precompiled is None:
                precompiled = compile_database()
                with tempfile.NamedTemporaryFile(
                            dir=os.path.dirname(database_path),
                            prefix=b'compiling.',
                            suffix=b'.db',
                            delete=False
                        ) as file:
                    file.write(pickle.dumps(precompiled))
                    file.flush()
                    file.close()
                    chmod_with_umask(file.name)
                    os.replace(file.name, database_path)

        if precompiled is None:
            return None
        return precompiled.data

    def _initialize_interrupt_handling(self) -> None:

        def handle_interrupt(signal_number: int, stack) -> None:
            revert_progress_changes()
            if parent_process() is None:
                log.info('Scan command interrupted, stopping...')
                self.terminate()
                reset_terminal()
            sys.exit(130)

        signal.signal(signal.SIGINT, handle_interrupt)

    def invoke(self) -> int:
        self._initialize_interrupt_handling()

        match_engine = MatchEngine.for_option(self.config.match_engine)
        pre_compile = (
                self.config.pre_compile
                or self.config.pre_compile_generic
            )
        signatures, database_source = self._get_signatures(
                match_engine,
                check_precompiled=not (
                        pre_compile or self.config.compile_local
                    )
            )
        match_engine_options = MatchEngineOptions(
                signature_set=signatures,
                pcre_options=self._get_pcre_options(),
                match_all=self.config.match_all,
            )
        if database_source is None:
            database_source = self._get_database_source(
                match_engine,
                match_engine_options,
                force=self.config.re_compile,
                generic=self.config.pre_compile_generic
            )
        match_engine_options.database_source = database_source

        if pre_compile:
            if match_engine_options.database_source is None:
                log.error(
                        'Signature set pre-compilation is not supported '
                        'with the current options'
                    )
                return 1
            else:
                log.info('Signature set successfully pre-compiled')
                return 0

        report_manager = ScanReportManager(
                self.context,
                signatures
            )
        io_manager = report_manager.get_io_manager()

        progress = None
        if self.config.progress:
            global screen_handler
            progress = ProgressDisplay(int(self.config.workers))
            screen_handler = progress.get_log_handler()
            if sys.stderr is None or sys.stderr.isatty():
                remove_initial_handler()
            log.addHandler(screen_handler)
            report_manager.set_progress_display(progress)

        paths = set()
        for argument in self.config.trailing_arguments:
            paths.add(argument)

        options = scanning.scanner.Options(
                paths=paths,
                match_engine_options=match_engine_options,
                workers=int(self.config.workers),
                chunk_size=self.config.chunk_size,
                scanned_content_limit=int(self.config.scanned_content_limit),
                file_filter=self._initialize_file_filter(),
                allow_io_errors=self.config.allow_io_errors,
                debug=self.config.debug,
                logging_initializer=self.context.get_log_settings().apply,
                match_engine=match_engine,
                profile=self.config.profile,
                profile_path=self.config.profile_path,
                direct_io=self.config.direct_io
            )
        if io_manager.should_read_stdin():
            options.path_source = io_manager.get_input_reader()

        with report_manager.open_output_file() as output_file:
            report = report_manager.initialize_report(
                    output_file
                )
            self.scanner = scanning.scanner.Scanner(options)
            if progress:
                use_log_events = True
            else:
                use_log_events = False
            metrics, timer = self.scanner.scan(
                    report.add_result,
                    progress.handle_update if progress is not None else None,
                    progress.scan_finished_handler if progress is not None
                    else None,
                    use_log_events
                )

            if progress:
                progress.end_on_input()
                revert_progress_changes()
                if progress.results_message:
                    print(progress.results_message)
            report.metrics = metrics
            report.timer = timer
            report.complete()
        return 0

    def terminate(self) -> None:
        if hasattr(self, 'scanner') and self.scanner is not None:
            self.scanner.terminate()
        reset_terminal()

    def generate_exception_message(
                self,
                exception: BaseException
            ) -> Optional[str]:
        if isinstance(exception, ProgressException):
            return (
                    'The current terminal is too small to '
                    'display progress output with the current scan '
                    'options'
                )
        return super().generate_exception_message(exception)


factory = MalwareScanSubcommand



from ..subcommands import SubcommandDefinition, UsageExample
from ..config.typing import ConfigDefinitions

config_definitions: ConfigDefinitions = {
    "overwrite": {
        "short_name": "o",
        "description": "Overwrite any existing configuration file without "
                       "prompting",
        "context": "CLI",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "request-license": {
        "short_name": "r",
        "description": "Automatically request a free licenses without "
                       "prompting",
        "context": "CLI",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "workers": {
        "short_name": "w",
        "description": "Specify the number of worker processes to "
                       "use for malware scanning",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": None,
        "meta": {
            "value_type": int
        }
    },
    "default": {
        "short_name": "D",
        "description": "Automatically accept the default values for any "
                       "options that are not explicitly specified. This will "
                       "also result in a free license being requested when "
                       "terms are accepted.",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False
    }
}

cacheable_types = set()

examples = [
    UsageExample(
        'Interactively configure Wordfence CLI',
        'wordfence configure'
    ),
    UsageExample(
        'Non-interactively configure Wordfence CLI to use 4 worker processes '
        'and the default values for all other options, automatically '
        'accepting the terms and requesting a free license',
        'wordfence configure --default --workers 4 --accept-terms'
    )
]

definition = SubcommandDefinition(
    name='configure',
    usage='',
    description='Configure Wordfence CLI',
    config_definitions=config_definitions,
    config_section='DEFAULT',
    cacheable_types=cacheable_types,
    requires_config=False,
    examples=examples
)



import os
from ...wordpress.site import WordpressLocator
from ...logging import log
from ..subcommands import Subcommand
from ..io import IoManager
from ..exceptions import ConfigurationException


class CountSitesSubcommand(Subcommand):

    def count_sites(self, path: bytes) -> int:
        count = 0
        locator = WordpressLocator(
                    path=path,
                    allow_nested=self.config.allow_nested,
                    allow_io_errors=self.config.allow_io_errors
                )
        for core in locator.locate_core_paths():
            log.debug('Located WordPress site at ' + os.fsdecode(core))
            count += 1
        return count

    def invoke(self) -> int:
        count = 0
        paths_counted = 0
        io_manager = IoManager(
                self.config.read_stdin,
                self.config.path_separator,
                binary=True
            )
        for path in self.config.trailing_arguments:
            count += self.count_sites(path)
            paths_counted += 1
        if io_manager.should_read_stdin():
            for path in io_manager.get_input_reader().read_all_entries():
                count += self.count_sites(path)
                paths_counted += 1
        if self.context.requires_input(self.config.require_path) \
                and paths_counted == 0:
            raise ConfigurationException(
                    'At least one path must be specified'
                )
        log.info(f'Located {count} WordPress site(s)')
        print(count)
        return 0


factory = CountSitesSubcommand
